
__all__ = [
    "App",
    "SettingsPane",
    "Manipulator",
]

import types, typing

import numpy as np

from ..JHTML import JHTML, DefaultOutputArea#, HTML, ActiveHTMLWrapper
from .Interfaces import *
from .Controls import Control, FunctionDisplay
from .Variables import WidgetControl, InterfaceVars, DefaultVars

# NonPrimitiveInterfaceType = typing.Union[
#     HTMLableType,
#     WidgetableType,
#     Component,
#     typing.Tuple[
#         typing.Union[
#             HTMLableType,
#             WidgetableType,
#             Component
#         ],
#         typing.Mapping
#     ]
# ]


__reload_hook__ = ['..JHTML', '.Interfaces', '.Controls', '.Variables']

# class FunctionDisplay:
#     def __init__(self, func, vars):
#         self.func = func
#         self.vars = vars
#         self.output = JupyterAPIs.get_widgets_api().Output()
#         self.update = self.update # a weird kluge to prevent a weakref issue...
#     def display(self):
#         self.output.clear_output()
#         with self.output:
#             res = self.func(**{x.name: x.value for x in self.vars})
#             if res is not None:
#                 JupyterAPIs.get_display_api().display(res)
#     def update(self, *ignored_settings):
#         return self.display()
#     def to_widget(self):
#         for v in self.vars:
#             v.callbacks.add(self.update)
#         return self.output
class Manipulator(Card):

    theme = Card.merge_themes(
        Card.theme,
        {
            'controls': {},
            'output': {}
        }
    )
    def __init__(self, func, *controls, debounce=None, autoclear=True, **etc):
        super().__init__(**etc)
        self.controls = [self.canonicalize_control(c) for c in controls]
        vars = [c.var for c in self.controls]
        self.output = FunctionDisplay(func, vars, debounce=debounce, autoclear=autoclear, **self.theme.get('output', {}))
        body = Flex(
            [
                self.output,
                Flex(self.controls, direction='column', **self.theme.get('controls', {}))
            ],
            direction='column'
        )
        self.component_args['body'] = (body,)
    @classmethod
    def canonicalize_control(cls, settings):
        if isinstance(settings, (WidgetControl, Control)):
            return settings
        else:
            var, settings = settings
            if not isinstance(settings, dict):
                if isinstance(settings, (list, tuple)) and isinstance(settings[0], (int, float, np.integer, np.floating)):
                    settings = {'range':settings}
                else:
                    settings = {'value':settings}
            # try:
            control = Control.construct(var, **settings)
            # except:
            #     control = WidgetControl(var, **settings)

            return control
    def initialize(self):
        self.output.update(None)

class App(Component):
    """
    Provides a framework for making Jupyter Apps with the
    elements built out in the Interfaces package
    """
    _app_stack=[]
    themes = {
        'primary':{
            'header':{
                'wrapper': {'cls': ['navbar-dark', 'bg-secondary', 'border-bottom']}
            },
            'toolbar':{'cls':['form-check', 'bg-light', 'border-bottom']},
            'sidebar':{
                'classes':['bg-light', 'border-end', 'h-100'],
                'styles':{},
                'opener':{
                    'header':{
                        'cls':['pt-0', 'bg-light'],
                        'border_top':'1px solid rgb(240, 240, 240)'
                    },
                    'body': {
                        'cls':['ps-0', 'pe-0']
                    }
                },
                'body':{
                    'wrapper':{'cls':['bg-light', 'border-end', 'h-100']}
                }
            },
            'footer':{
                 'wrapper': {'cls': ['navbar-light', 'bg-light', 'border-top']}
            }
        }
    }
    @classmethod
    def merge_themes(cls, theme_1, theme_2):
        new = theme_1.copy()
        for k in theme_2:
            if k in new and isinstance(new[k], dict):
                new[k] = cls.merge_themes(new[k], theme_2[k])
            else:
                new[k] = theme_2[k]
        return new


    def __init__(self,
                 body=None,
                 header=None,
                 footer=None,
                 sidebar=None,
                 toolbar=None,
                 theme='primary',
                 layout='grid',
                 cls='app border',
                 output=None,
                 capture_output=None,
                 vars=None,
                 **attrs
                 ):
        self._parent = None if len(self._app_stack) == 0 else self._app_stack[-1]
        super().__init__()
        self.vars = DefaultVars.resolve() if vars is None else vars
        self.output = JHTML.OutputArea(autoclear=True) if output is None else output
        self.capture_output = self._parent is None if capture_output is None else capture_output
        self.theme = self.themes[theme] if isinstance(theme, str) else self.merge_themes(self.themes['primary'], theme)
        self._body = [None, body]
        self._header = [None, header]
        self._footer = [None, footer]
        self._sidebar = [None, sidebar]
        self._toolbar = [None, toolbar]
        self.layout = layout
        self.cls = cls
        self.attrs = attrs
        self._out = None
        self._vv = None
        self._stack_depth = 0
    def __enter__(self):
        if self._stack_depth == 0:
            self._vv = DefaultVars(self.vars)
            self._vv.__enter__()
            self._out = DefaultOutputArea(self.output)
            self._out.__enter__()
            self._app_stack.append(self)
        self._stack_depth += 1
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stack_depth -= max(0, self._stack_depth - 1)
        if self._stack_depth == 0:
            self._vv.__exit__(exc_type, exc_val, exc_tb)
            self._vv = None
            self._out.__exit__(exc_type, exc_val, exc_tb)
            self._out = None
            self._app_stack.remove(self)
    @property
    def body(self):
        if self._body[1] is not None:
            if self._body[0] is None:
                with self:
                    self._body[0] = self.construct_body(self._body[1])
        return self._body[0]
    @body.setter
    def body(self, b):
        self._body = [None, b]
    @property
    def header(self):
        if self._header[1] is not None:
            if self._header[0] is None:
                with self:
                    self._header[0] = self.construct_header(self._header[1])
        return self._header[0]
    @header.setter
    def header(self, h):
        self._header = [None, h]
    @property
    def sidebar(self):
        if self._sidebar[1] is not None:
            if self._sidebar[0] is None:
                with self:
                    self._sidebar[0] = self.construct_sidebar(self._sidebar[1])
        return self._sidebar[0]
    @sidebar.setter
    def sidebar(self, s):
        self._sidebar = [None, s]
    @property
    def toolbar(self):
        if self._toolbar[1] is not None:
            if self._toolbar[0] is None:
                with self:
                    self._toolbar[0] = self.construct_toolbar(self._toolbar[1])
        return self._toolbar[0]
    @toolbar.setter
    def toolbar(self, t):
        self._toolbar = [None, t]
    @property
    def footer(self):
        if self._footer[1] is not None:
            if self._footer[0] is None:
                with self:
                    self._footer[0] = self.construct_footer(self._footer[1])
        return self._footer[0]
    @footer.setter
    def footer(self, f):
        self._footer = [None, f]
    @classmethod
    def prep_head_item(cls, item):
        if (
                isinstance(item, (tuple, list))
                and len(item) == 2
                and isinstance(item[1], (types.FunctionType, types.MethodType))
        ):
            item = Button(*item)
        return item
    @classmethod
    def construct_navbar_item(cls, item):
        if isinstance(item, dict) and len(item) == 1 and 'body' not in item:
            for item in item.items():
                item = tuple(item)
        if isinstance(item, (tuple, list)) and len(item) == 2:
            k,v = item
            if isinstance(v, (tuple, list, dict)):
                v = tuple(cls.prep_head_item(i) for i in v)
                item = {'raw':Dropdown(k, v)}
            elif isinstance(v, (types.FunctionType, types.MethodType)):
                item = Button(k, v)
        return item
    def construct_header(self, header, **opts):
        if isinstance(header, tuple) and len(header) == 2 and isinstance(header[1], dict):
            opts = dict(header[1], **opts)
            header = header[0]
        elif isinstance(header, dict):
            header = header.copy()
            sb = header['items']
            del header['items']
            opts = dict(header, **opts)
            header = sb
        elif not (isinstance(header, (list, tuple, Component)) or hasattr(header, 'to_widget') or hasattr(header, 'to_tree')):
            header = [header]
        if isinstance(header, (list, tuple)):
            header = Navbar(
                [self.construct_navbar_item(h) for h in header],
                **dict({'theme':self.theme['header']}, **opts)
            )
        # elif isinstance(header, str):
        #     header = ...
        # elif isinstance(header, (HTML.XMLElement, ActiveHTMLWrapper)):
        #     ...
        return header
    def construct_footer(self, footer, **opts):
        if isinstance(footer, tuple) and len(footer) == 2 and isinstance(footer[1], dict):
            opts = dict(footer[1], **opts)
            footer = footer[0]
        elif isinstance(footer, dict):
            footer = footer.copy()
            sb = footer['items']
            del footer['items']
            opts = dict(footer, **opts)
            footer = sb
        elif not (isinstance(footer, (list, tuple, Component)) or hasattr(footer, 'to_widget') or hasattr(footer, 'to_tree')):
            footer = [footer]
        if isinstance(footer, (list, tuple)):
            footer = Navbar(
                footer,
                **dict({'theme':self.theme['footer']}, **opts)
            )
        return footer
    def construct_sidebar_item(self, item):
        if isinstance(item, tuple):
            if isinstance(item[0], tuple) and len(item[0]) == 2:
                items = []
                for k,v in item:
                    if isinstance(v, tuple):
                        v = Sidebar([self.construct_sidebar_item(h) for h in v])
                    items.append((k, v))
                item = dict(body=Opener(items, theme=self.theme['sidebar']['opener']['header']), theme=self.theme['sidebar']['opener']['body'])
            else:
                k,v = item
                if isinstance(v, tuple):
                    v = Sidebar([self.construct_sidebar_item(h) for h in v])
                item = dict(body=Opener(((k, v),), theme=self.theme['sidebar']['opener']['header']), theme=self.theme['sidebar']['opener']['body'])
        return item
    def construct_sidebar(self, sidebar, **opts):
        if isinstance(sidebar, tuple) and len(sidebar) == 2 and isinstance(sidebar[1], dict):
            opts = dict(sidebar[1], **opts)
            sidebar = sidebar[0]
        elif isinstance(sidebar, dict):
            sidebar = sidebar.copy()
            sb = sidebar['items']
            del sidebar['items']
            opts = dict(sidebar, **opts)
            sidebar = sb
        elif not (isinstance(sidebar, (list, tuple, Component)) or hasattr(sidebar, 'to_widget') or hasattr(sidebar, 'to_tree')):
            sidebar = [sidebar]
        if isinstance(sidebar, (list, tuple)):
            sidebar = Sidebar(
                [self.construct_sidebar_item(h) for h in sidebar],
                **dict({'theme':self.theme['sidebar']['body']}, **opts)
            )
        return sidebar

    def construct_toolbar_item(self, item):
        if isinstance(item, dict):
            item = item.copy()
            var = item['var']
            del item['var']
            item = Control.construct(var, **item)
        return item
    def construct_toolbar(self, toolbar, **opts):
        if not (isinstance(toolbar, (list, tuple, Component)) or hasattr(toolbar, 'to_widget') or hasattr(toolbar, 'to_tree')):
            toolbar = [toolbar]
        if isinstance(toolbar, (list, tuple)):
            if isinstance(toolbar[0], (list, tuple)):
                toolbar = Grid(
                    [[self.construct_toolbar_item(h) for h in row] for row in toolbar],
                    **dict(self.theme['toolbar'], **opts)
                )
            else:
                toolbar = JHTML.Div(
                    [self.construct_toolbar_item(h) for h in toolbar],
                    **dict(self.theme['toolbar'], **opts)
                )
        return toolbar

    def wrap_body(self, fn, **styles):
        # @functools.wraps(fn)
        # def wrapper(event=None, pane=None, **vars):
        #     return fn(event=event, pane=pane, **vars)
        return FunctionDisplay(fn, self.vars, **styles)
    def construct_body_item(self, item):
        if isinstance(item, (types.MethodType, types.FunctionType, types.LambdaType)):
            item = self.wrap_body(item)
        elif not isinstance(item, (str, Component, JHTML.Styled)) and not (hasattr(item, 'to_string') or hasattr(item, 'to_jhtml')):
            try:
                item, styles = item
            except (ValueError, TypeError):
                pass
            else:
                if isinstance(item, (types.MethodType, types.FunctionType, types.LambdaType)):
                    item = self.wrap_body(item, **styles)
                else:
                    item = JHTML.Span(item, **styles)
        return item
    def construct_body(self, body):
        if isinstance(body, dict):
            body = Tabs({k:self.construct_body_item(b) for k,b in body.items()})
        elif isinstance(body, (list, tuple)):
            body = [self.construct_body(b) for b in body]
        else:
            body = self.construct_body_item(body)
        return body

    def construct_layout(self):
        with self:
            if self.layout == 'grid':
                elements = []
                nrows = 0
                ncols = 0

                # if self.header is not None:
                #     nrows += 1
                if self.sidebar is not None:
                    ncols += 1
                if self.toolbar is not None:
                    nrows += 1
                    ncols += 1
                if self.capture_output:
                    nrows += 1
                if self.body is not None:
                    nrows += 1 if not isinstance(self.body, list) else len(self.body)
                    ncols = min(ncols+1, 2)
                # if self.footer is not None:
                #     nrows += 1

                column_width = []
                row_height = []
                if self.header is not None:
                    header = []
                    header.append(Grid.Item(self.header, col_span=ncols))
                    elements.append(header)
                    row_height.append('auto')

                body = []
                if self.sidebar is not None:
                    body.append(Grid.Item(self.sidebar, row_span=nrows))
                    column_width.append('auto')
                if self.toolbar is not None:
                    body.append(self.toolbar)
                    elements.append(body)
                    if self.sidebar is not None:
                        body = [None]
                    else:
                        body = []
                    row_height.append('auto')
                if isinstance(self.body, (list, tuple)):
                    for i,b in enumerate(self.body):
                        body.append(b)
                        if i == 0:
                            row_height.append('1fr')
                            column_width.append('1fr')
                        else:
                            row_height.append('auto')
                        elements.append(body)
                        if self.sidebar is not None:
                            body = [None]
                        else:
                            body = []
                else:
                    body.append(self.body)
                    row_height.append('1fr')
                    column_width.append('1fr')
                    elements.append(body)

                if self.capture_output:
                    output_list = []
                    if self.sidebar is not None:
                        output_list.append(None)
                    output_list.append(Grid.Item(self.output))
                    elements.append(output_list)
                    row_height.append('auto')

                if self.footer is not None:
                    footer = []
                    footer.append(Grid.Item(self.footer, col_span=ncols))
                    elements.append(footer)
                    row_height.append('auto')

                layout = Grid(
                    elements,
                    column_width=column_width,
                    row_height=row_height,
                    cls=self.cls,
                    **self.attrs
                )
            else:
                raise NotImplementedError("ugh")
        return layout

    def to_jhtml(self):
        return self.construct_layout().to_jhtml()
class SettingsPane(App):
    themes = {
        'primary': App.merge_themes(
            App.themes['primary'],
            {'toolbar': dict(cls=['form-check'])}
        )
    }
    def __init__(self, settings, cls=None, **opts):
        super().__init__(
            toolbar=settings,
            cls=cls,
            **opts
        )



import ast, collections, inspect

__all__ = [
    "ExamplesParser"
]

class ExamplesParser:
    """
    Provides a parser for unit tests to turn them into examples
    """

    def __init__(self, unit_tests):
        self.source = unit_tests
        self.ast = ast.parse(unit_tests)
        self._class_node = None
        self._setup = None
        self._functions = None
        self._functions_map = None
        self._node_cache = {}

    def find_setup(self, tree_iter):
        setup = []
        for node in tree_iter:
            node_type = type(node).__name__
            if node_type not in ["ClassDef"]:
                setup.append(node)
            else:
                break
        else:
            node = None
        return node, setup

    def parse_tests(self, tree_iter):
        """
        Parses out the
        :param tree_iter:
        :type tree_iter:
        :return:
        :rtype:
        """
        class_setup = []
        functions = collections.OrderedDict()
        for node in tree_iter:
            node_type = type(node).__name__
            if node_type == 'FunctionDef':
                fname = node.name
                if fname.startswith('test_'):
                    functions[fname.split("_", 1)[1]] = node
                else:
                    class_setup.append(node)
            else:
                class_setup.append(node)
            # else:
            #     raise ValueError(
            #         "AST node of type {} with body {} not handled".format(
            #             node_type, ast.get_source_segment(self.source, node)
            #         )
            #     )
        return class_setup, functions

    def walk_tree(self):
        if hasattr(self.ast, 'body'):
            tree = self.ast.body
        else:
            tree = self.ast
        tree_iter = iter(tree)
        class_node, base_setup = self.find_setup(tree_iter)
        class_setup, functions = self.parse_tests(class_node.body)
        self._class_node = class_node
        self._setup = (base_setup, class_setup)
        self._functions = functions
        self._functions_map = self.load_function_map()
        return self._setup, self._functions

    def format_node(self, node):
        try:
            repr = self._node_cache[node]
        except KeyError:
            base = ast.get_source_segment(self.source, node)
            indent = " "*node.col_offset
            repr = indent + base
            self._node_cache[node] = repr
        return repr

    @classmethod
    def from_file(cls, tests_file):
        with open(tests_file) as f:
            return cls(f.read())

    @property
    def class_spec(self):
        if self._class_node is None:
            self.walk_tree()
        return (self._class_node, self._setup[1])
    @property
    def setup(self):
        if self._setup is None:
            self.walk_tree()
        return self._setup[0]
    @property
    def functions(self):
        if self._functions is None:
            self.walk_tree()
        return self._functions
    @property
    def functions_map(self):
        if self._functions is None:
            self.walk_tree()
        return self._functions_map

    def load_function_map(self):
        mapping = {}
        for k,v in self._functions.items():
            for f in self.get_examples_functions(v):
                if f not in mapping:
                    mapping[f] = [k]
                else:
                    mapping[f].append(k)
        return mapping
    def _handle_stmt(self, stmt, all_fns):
        if hasattr(stmt, 'body'):
            all_fns.update(self.get_examples_functions(stmt))
        elif hasattr(stmt, 'names'):
            for name in stmt.names:
                if isinstance(name, str):
                    all_fns.add(name)
                else:
                    all_fns.add(name.name)
        elif hasattr(stmt, 'id'):
            name = stmt.id
            if isinstance(name, str):
                all_fns.add(name)
            else:
                all_fns.add(name.name)
        elif hasattr(stmt, 'func'):
            self._handle_stmt(stmt.func, all_fns)
            # for substmt in stmt.args:
            #     self._handle_stmt(substmt, all_fns)
        # we'll only check common patterns for speed purposes...
        elif hasattr(stmt, 'operand'):
            pass
            # self._handle_stmt(stmt.operand, all_fns)
        elif hasattr(stmt, 'right'):
            pass
            # self._handle_stmt(stmt.left, all_fns)
            # self._handle_stmt(stmt.right, all_fns)
        elif hasattr(stmt, 'comparators'):
            pass
            # self._handle_stmt(stmt.left, all_fns)
            # for v in stmt.comparators:
            #     self._handle_stmt(v, all_fns)
        elif hasattr(stmt, 'value'):
            if hasattr(stmt, 'targets'):
                # assignment
                self._handle_stmt(stmt.value, all_fns)
            else:
                node_type = type(stmt).__name__
                if node_type == 'Constant':
                    pass
                else:
                    self._handle_stmt(stmt.value, all_fns)
        elif hasattr(stmt, 'values'):
            pass
            # for v in stmt.values:
            #     self._handle_stmt(v, all_fns)
        elif hasattr(stmt, 'elts'):
            pass
            # for v in stmt.elts:
            #     self._handle_stmt(v, all_fns)
        elif (
            hasattr(stmt, 'generators')
            or hasattr(stmt, 'exc')
            or hasattr(stmt, 'targets')
        ):
            pass # for now at least...
        else:
            raise Exception(stmt, stmt._fields)
    def get_examples_functions(self, node):
        all_fns = set()
        try:
            for stmt in node.body:
                self._handle_stmt(stmt, all_fns)
        except TypeError:
            self._handle_stmt(node.body, all_fns)
        return all_fns

    def filter_by_name(self, name):
        import copy
        if self._functions is None:
            self.walk_tree()
        c = copy.copy(self)
        try:
            keys = c._functions_map[name]
        except KeyError:
            new_fns = {}
        else:
            new_fns = {k:c._functions[k] for k in keys if k in c._functions}
        if len(new_fns) == 0:
            return None
        else:
            c._functions = new_fns
            return c


"""
Provides support for chemical toolkits
"""

from .Interface import *

__all__ = [
    "OpenBabelInterface",
    "PybelInterface",
    "RDKitInterface",
    "OpenChemistryInterface",
    "CCLibInterface"
]

class OpenBabelInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the OpenBabel toolkit (which is installed with anaconda)
    """
    name = 'OpenBabel'
    module = 'openbabel.openbabel'

class PybelInterface(ExternalProgramInterface):
    """
    A simple class to support operations that make use of the OpenBabel toolkit (which is installed with anaconda)
    """
    name = 'Pybel'
    module = 'openbabel.pybel'

class RDKitInterface(ExternalProgramInterface):
    name = 'RDKit'
    module = 'rdkit'

class OpenChemistryInterface:
    name = 'OpenChemistry'
    module = 'openchemistry.io'

# PYBEL_SUPPORTED = None
    # OB_SUPPORTED = None
    # def __init__(self):
    #     self._lib = None
    # @classmethod
    # def _pybel_installed(cls):
    #     if cls.PYBEL_SUPPORTED is None:
    #         try:
    #             import openbabel.pybel
    #         except ImportError:
    #             cls.PYBEL_SUPPORTED = False
    #         else:
    #             cls.PYBEL_SUPPORTED = True
    #
    #     return cls.PYBEL_SUPPORTED
    # @classmethod
    # def _ob_installed(cls):
    #     if cls.OB_SUPPORTED is None:
    #         try:
    #             import openbabel.openbabel
    #         except ImportError:
    #             cls.OB_SUPPORTED = False
    #         else:
    #             cls.OB_SUPPORTED = True
    #
    #     return cls.OB_SUPPORTED
    # @property
    # def pybel(self):
    #     if self._pybel is None:
    #         if self._pybel_installed():
    #             import openbabel.pybel as lib
    #             self._pybel = lib
    #         else:
    #             raise ImportError("OpenBabel isn't installed")
    #     return self._pybel
    # @property
    # def openbabel(self):
    #     if self._ob is None:
    #         if self._ob_installed():
    #             import openbabel.openbabel as lib
    #             self._ob = lib
    #         else:
    #             raise ImportError("OpenBabel isn't installed")
    #     return self._ob


class CCLibInterface:
    name = 'CCLib'
    module = 'cclib'


"""
Provides a uniform interface for potentially installed external programs
"""

__all__ = [
    "ExternalProgramInterface"
]

import weakref
import importlib


class ExternalProgramInterface:
    name = None
    module = None
    lib_supported = None
    library = None

    @classmethod
    def try_load_lib(cls):
        if cls.library is None:
            if cls.lib_supported is False:
                raise ImportError("Library '{}' is not installed".format(cls.name))
            try:
                cls.library = cls.load_library()
            except ImportError:
                cls.lib_supported = False
            else:
                cls.lib_supported = True
        return cls.lib_supported

    @classmethod
    def get_lib(cls):
        if cls.try_load_lib():
            return cls.library
        else:
            raise ImportError("module {} not installed".format(cls.module))

    @classmethod
    def load_library(cls):
        return __import__(cls.module)
        # raise NotImplementedError("{} needs to implement `load_library`".format(cls.__name__))

    method_table = weakref.WeakValueDictionary()
    @classmethod
    def method(cls, name):
        if name not in cls.method_table:
            cls.method_table[name] = getattr(cls.get_lib(), name)
        return cls.method_table[name]
    @classmethod
    def submodule(cls, submodule):
        check_load = cls.get_lib()
        return importlib.import_module(cls.module+"."+submodule)
    @property
    def lib(self):
        return self.get_lib()
    def __getattr__(self, item):
        return self.method(item)


"""
Handles all the nastiness of Matplotlib properties so that we can use a more classically python plotting method
"""

from .Backends import DPI_SCALING
from .Styling import Styled, PlotLegend

__all__ = [
    "GraphicsPropertyManager",
    "GraphicsPropertyManager3D"
]

__reload_hook__ = [".Styling"]

class GraphicsPropertyManager:
    """
    Manages properties for Graphics objects so that concrete GraphicsBase instances don't need to duplicate code, but
    at the same time things that build off of GraphicsBase don't need to implement all of these properties
    """

    def __init__(self, graphics, figure, axes, managed=False):
        self.graphics = graphics
        self.managed = managed # if there's an external manager
        self.figure = figure
        self.axes = axes
        self._figure_label = None
        self._plot_label = None
        self._plot_legend = None
        self._legend_style = None
        self._axes_labels = None
        self._frame = None
        self._frame_style = None
        self._plot_range = None
        self._ticks = None
        self._scale = None
        self._ticks_style = None
        self._ticks_label_style = None
        self._aspect_ratio = None
        self._image_size = None
        self._padding = None
        self._background = None
        self._colorbar = None
        self._cbar_obj = None
        self._spacings = None

    @property
    def figure_label(self):
        return self._figure_label
    @figure_label.setter
    def figure_label(self, label):
        self._figure_label = label
        if not self.managed:
            if label is None:
                self.figure.set_figure_label("")
            elif isinstance(label, Styled):
                self.figure.set_figure_label(*label.val, **label.opts)
            else:
                self.figure.set_figure_label(label)

    @property
    def plot_label(self):
        return self._plot_label
    @plot_label.setter
    def plot_label(self, label):
        self._plot_label = label
        if label is None:
            self.axes.set_plot_label("")
        elif isinstance(label, Styled):
            self.axes.set_plot_label(*label.val, **label.opts)
        elif Styled.could_be(label):
            label = Styled.construct(label)
            self.axes.set_plot_label(*label.val, **label.opts)
        else:
            self.axes.set_plot_label(label)

    # set plot legend
    @property
    def plot_legend(self):
        return self._plot_legend
    @plot_legend.setter
    def plot_legend(self, legend):
        self._plot_legend = legend

        if PlotLegend.could_be_legend(legend):
            self._plot_legend = PlotLegend.construct(legend)
        else:
            raise NotImplementedError("plot legends not properly supported..")
            artists = self.graphics.artists
            if artists is not None:
                if legend is None:
                    for a in artists:
                        a.remove_label("")
                elif legend is True:
                    pass
                elif isinstance(legend, Styled):
                    for a in artists:
                        a.set_label(legend.val, **legend.opts)
                elif Styled.could_be(legend):
                    legend = Styled.construct(legend)
                    for a in artists:
                        a.set_label(legend.val, **legend.opts)
                else:
                    for a in artists:
                        a.set_label(legend)

    @property
    def legend_style(self):
        return self._legend_style
    @legend_style.setter
    def legend_style(self, style):
        self._legend_style = style


    # # set axes labels
    # @property
    # def axes(self):
    #     return self._axes
    # @axes.setter
    # def axes(self, axes):
    #     ...

    # set axes labels
    @property
    def axes_labels(self):
        return self._axes_labels
    @axes_labels.setter
    def axes_labels(self, labels):
        if self._axes_labels is None:
            self._axes_labels = (self.axes.get_xlabel(), self.axes.get_ylabel())
        try:
            xlab, ylab = labels
        except ValueError:
            xlab, ylab = labels = (labels, self._axes_labels[1])

        self._axes_labels = tuple(labels)
        if xlab is None:
            self.axes.set_xlabel("")
        elif isinstance(xlab, Styled):
            self.axes.set_xlabel(xlab.val[0], **xlab.opts)
        elif Styled.could_be(xlab):
            xlab = Styled.construct(xlab)
            self.axes.set_xlabel(*xlab.val, **xlab.opts)
        else:
            self.axes.set_xlabel(xlab)
        if ylab is None:
            self.axes.set_ylabel("")
        elif isinstance(ylab, Styled):
            self.axes.set_ylabel(*ylab.val, **ylab.opts)
        elif Styled.could_be(ylab):
            ylab = Styled.construct(ylab)
            self.axes.set_ylabel(*ylab.val, **ylab.opts)
        else:
            self.axes.set_ylabel(ylab)

    # set plot ranges
    @property
    def plot_range(self):
        if self._plot_range is None:
            pr = (list(sorted(self.axes.get_xlim())), list(sorted(self.axes.get_ylim())))
        else:
            pr = self._plot_range
        return pr
    @plot_range.setter
    def plot_range(self, ranges):
        if self._plot_range is None:
            self._plot_range = (self.axes.get_xlim(), self.axes.get_ylim())
        try:
            x, y = ranges
        except ValueError:
            x, y = ranges = (self._plot_range[0], ranges)
        else:
            if isinstance(x, int) or isinstance(x, float):
                x, y = ranges = (self._plot_range[0], ranges)

        self._plot_range = tuple(ranges)

        if isinstance(x, Styled):  # name feels wrong here...
            self.axes.set_xlim(*x.val, **x.opts)
        elif Styled.could_be(x):
            x = Styled.construct(x)
            self.axes.set_xlim(*x.val, **x.opts)
        elif x is not None:
            self.axes.set_xlim(x)
        if isinstance(y, Styled):
            self.axes.set_ylim(*y.val, **y.opts)
        elif Styled.could_be(y):
            y = Styled.construct(y)
            self.axes.set_ylim(*y.val, **y.opts)
        elif y is not None:
            self.axes.set_ylim(y)

    # set plot ticks
    @property
    def ticks(self):
        return self._ticks
    def _set_ticks(self, x, set_ticks=None, set_locator=None, set_minor_locator=None, **opts):
        import matplotlib.ticker as ticks

        if isinstance(x, Styled):
            self._set_ticks(*x.val,
                            set_ticks=set_ticks,
                            set_locator=set_locator, set_minor_locator=set_minor_locator,
                            **x.opts
                            )
        elif Styled.could_be(x):
            x = Styled.construct(x)
            self._set_ticks(*x.val,
                            set_ticks=set_ticks,
                            set_locator=set_locator, set_minor_locator=set_minor_locator,
                            **x.opts
                            )
        elif isinstance(x, ticks.Locator):
            set_locator(x)
        elif isinstance(x, (list, tuple)):
            if len(x) == 2 and isinstance(x[0], ticks.Locator):
                set_locator(x[0])
                set_minor_locator(x[1])
            if len(x) == 2 and isinstance(x[0], (list, tuple)):
                set_ticks(*x, **opts)
            else:
                set_ticks(x, **opts)
        elif isinstance(x, (float, int)):
            set_ticks(ticks.MultipleLocator(x), **opts)
        elif x is not None:
            set_ticks(x, **opts)
    def _set_xticks(self, x, **opts):
        return self._set_ticks(x,
                               set_ticks=self.axes.set_xticks,
                               set_locator=self.axes.xaxis.set_major_locator,
                               set_minor_locator=self.axes.xaxis.set_minor_locator,
                               **opts
                               )

    def _set_yticks(self, y, **opts):
        return self._set_ticks(y,
                               set_ticks=self.axes.set_yticks,
                               set_locator=self.axes.yaxis.set_major_locator,
                               set_minor_locator=self.axes.yaxis.set_minor_locator,
                               **opts
                               )

    @ticks.setter
    def ticks(self, ticks):
        if self._ticks is None:
            self._ticks = (self.axes.get_xticks(), self.axes.get_yticks())
        try:
            x, y = ticks
        except (ValueError, TypeError):
            if isinstance(ticks, bool):
                x, y = ticks = (ticks, ticks)
            else:
                x, y = ticks = (self._ticks[0], ticks)

        self._ticks = ticks
        self._set_xticks(x)
        self._set_yticks(y)

    @property
    def ticks_style(self):
        return self._ticks_style
    @ticks_style.setter
    def ticks_style(self, ticks_style):
        if self._ticks_style is None:
            self._ticks_style = (None,) * 2
        if isinstance(ticks_style, dict):
            ticks_style = (ticks_style, ticks_style)
        try:
            x, y = ticks_style
        except ValueError:
            x, y = ticks_style = (ticks_style, ticks_style)
        self._ticks_style = ticks_style
        if x is not None:
            if x is True:
                x = dict(bottom=True, labelbottom=True)
            elif x is False:
                x = dict(bottom=False, top=False, labelbottom=False, labeltop=False)
            self.axes.set_xtick_style(
                **x
            )
        if y is not None:
            if y is True:
                y = dict(left=True, labelleft=True)
            elif y is False:
                y = dict(left=False, right=False, labelleft=False, labelright=False)
            self.axes.set_ytick_style(
                **y
            )
    @property
    def frame_style(self):
        return self._frame_style
    @frame_style.setter
    def frame_style(self, f_style):
        if self._frame_style is None:
            self._frame_style = ((None, None), (None, None))
        if isinstance(f_style, dict):
            f_style = ((f_style, f_style), (f_style, f_style))

        try:
            x, y = f_style
        except ValueError:
            x, y = f_style = (f_style, f_style)
        if isinstance(y, dict):
            y = (y, y)
        if isinstance(x, dict):
            x = (x, x)
        if len(y) == 2:
            b, t = y
        else:
            b = t = y
        if len(x) == 2:
            l, r = x
        else:
            l = r = x

        self._frame_style = ((l, r), (b, t))
        self.axes.set_frame_style(
            self._frame_style
        )

    ticks_label_base_styles = {
        'size', 'color',
        'top', 'left', 'right', 'bottom'
    }
    ticks_label_style_remapping={'fontsize':'size', 'fontcolor':'color'}
    @classmethod
    def clean_tick_label_styles(cls, k):
        if k.startswith('label'):
            k = k[5:]
        if k in cls.ticks_label_base_styles:
            k = 'label' + k
        k = cls.ticks_label_style_remapping.get(k, k)
        return k

    @property
    def ticks_label_style(self):
        return self._ticks_label_style
    @ticks_label_style.setter
    def ticks_label_style(self, ticks_style):
        if self._ticks_label_style is None:
            self._ticks_label_style = (None,) * 2
        try:
            x, y = ticks_style
        except ValueError:
            x, y = ticks_style = (ticks_style, ticks_style)
        self._ticks_label_style = ticks_style

        if x is not None:
            self.axes.set_xtick_style(
                **{self.clean_tick_label_styles(k):v for k,v in x.items()}
            )
        if y is not None:
            self.axes.set_ytick_style(
                **{self.clean_tick_label_styles(k):v for k,v in y.items()}
            )

    @property
    def aspect_ratio(self):
        return self._aspect_ratio
    @aspect_ratio.setter
    def aspect_ratio(self, ar):
        if isinstance(ar, (float, int)):
            a, b = self.plot_range
            cur_ar = abs(b[1] - b[0]) / abs(a[1] - a[0])
            targ_ar = ar / cur_ar
            self.axes.set_aspect_ratio(targ_ar)
        elif isinstance(ar, str):
            self.axes.set_aspect_ratio(ar)
        else:
            self.axes.set_aspect_ratio(ar[0], **ar[1])
        self._aspect_ratio = ar

    def _compute_inset_imagesize(self):
        if self.managed or self.graphics.inset:
            bbox = self.axes.get_bbox()
            ((alx, aby), (arx, aty)) = bbox
            ((plx, prx), (pby, pty)) = self.padding
            return (arx - alx + prx + plx, aty - aby + pty + pby)
        else:
            return tuple(s * DPI_SCALING for s in self.figure.get_size_inches())
    # set size
    @property
    def image_size(self):
        # im_size = self._image_size
        # if isinstance(self._image_size, (int, float)):
        #     im_size =
        if isinstance(self._image_size, str) and self._image_size == 'auto':
            return self._compute_inset_imagesize()
        else:
            return self._image_size
    @image_size.setter
    def image_size(self, wh):
        if isinstance(wh, str) and wh == 'auto':
            self._image_size = wh
        else:
            if self._image_size is None:
                self._image_size = self._compute_inset_imagesize()
            try:
                w, h = wh
            except (TypeError, ValueError):
                ar = self.aspect_ratio
                if not isinstance(ar, (int, float)):
                    try:
                        ar = self._image_size[1] / self._image_size[0]
                    except TypeError:
                        ar = 1
                w, h = wh = (wh, ar * wh)
            self._image_size = (w, h)

            if (
                    (w is not None or h is not None)
                    and not self.managed
                    and not self.graphics.inset
            ):
                if w is None:
                    w = self._image_size[0]
                if h is None:
                    h = self._image_size[1]

                if w > DPI_SCALING: # refuse to have anything smaller than 1 inch?
                    wi = w / DPI_SCALING
                else:
                    wi = w
                    w = DPI_SCALING * w

                if h > DPI_SCALING:
                    hi = h / DPI_SCALING
                else:
                    hi = h
                    h = DPI_SCALING * h
                self.figure.set_size_inches(wi, hi)
    @property
    def axes_bbox(self):
        return self.axes.get_bbox()
        # return bbox
    @axes_bbox.setter
    def axes_bbox(self, bbox):
        if bbox is not None:
            self.axes.set_bbox(bbox)

    # set background color
    @property
    def background(self):
        if self._background is None:
            if self.graphics.inset:
                self._background = self.axes.get_facecolor()
                self.fe_background = self.axes.get_facecolor()
            else:
                self._background = self.figure.get_facecolor()
        return self._background
    @background.setter
    def background(self, bg):
        self._background = bg
        if not self.managed and not self.graphics.inset:
            self.figure.set_facecolor(bg)
        self.axes.set_facecolor(bg)

    # set show_frame
    @property
    def frame(self):
        return self._frame
    @frame.setter
    def frame(self, fr):
        self._frame = fr
        if fr is True or fr is False:
            self.axes.set_frame_visible(fr)
        else:
            lr, bt = fr
            if len(lr) == 2:
                l, r = lr
            else:
                l = lr; r = lr
            if len(bt) == 2:
                b, t = bt
            else:
                b = bt; t = bt
            self.axes.set_frame_visible(
                ((l, r), (b, t))
            )

    @property
    def scale(self):
        return self._scale
    @scale.setter
    def scale(self, scales):
        if self._scale is None:
            self._scale = (self.axes.get_xscale(), self.axes.get_yscale())
        try:
            x, y = scales
        except ValueError:
            x, y = scales = (self._scale[0], scales)

        self._scale = tuple(scales)

        if isinstance(x, Styled):
            self.axes.set_xscale(*x.val, **x.opts)
        elif Styled.could_be(x):
            x = Styled.construct(x)
            self.axes.set_xscale(*x.val, **x.opts)
        elif x is not None:
            self.axes.set_xscale(x)
        if isinstance(y, Styled):
            self.axes.set_yscale(*y.val, **y.opts)
        elif y is not None:
            self.axes.set_yscale(y)
        elif Styled.could_be(y):
            y = Styled.construct(y)
            self.axes.set_yscale(*y.val, **y.opts)

    @property
    def padding(self):
        if self.managed or self.graphics.inset:
            return self.axes.get_padding()
        else:
            return self._padding
    @padding.setter
    def padding(self, padding):
        try:
            w, h = padding
        except (ValueError, TypeError):
            w = h = padding
        try:
            wx, wy = w
        except (ValueError, TypeError):
            wx = wy = w
        try:
            hx, hy = h
        except (ValueError, TypeError):
            hx = hy = h

        W, H = self.image_size
        if wx < 1:
            wx = round(wx * W)
        if wy < 1:
            wy = round(wy * W)
        if hx < 1:
            hx = round(hx * H)
        if hy < 1:
            hy = round(hy * H)
        self._padding = ((wx, wy), (hx, hy))
        wx = wx / W; wy = wy / W
        hx = hx / H; hy = hy / H
        if not self.managed and not self.graphics.inset:
            self.figure.set_extents([
                [wx, 1-wy],
                [hx, 1-hy]
            ])
    @property
    def padding_left(self):
        return self._padding[0][0]
    @padding_left.setter
    def padding_left(self, p):
        wx, wy = self._padding[0]
        hx, hy = self._padding[1]
        self.padding = ((p, wy), (hx, hy))
    @property
    def padding_right(self):
        return self._padding[0][1]
    @padding_right.setter
    def padding_right(self, p):
        wx, wy = self._padding[0]
        hx, hy = self._padding[1]
        self.padding = ((wx, p), (hx, hy))
    @property
    def padding_top(self):
        return self._padding[1][1]
    @padding_top.setter
    def padding_top(self, p):
        wx, wy = self._padding[0]
        hx, hy = self._padding[1]
        self.padding = ((wx, wy), (hx, p))
    @property
    def padding_bottom(self):
        return self._padding[1][0]
    @padding_bottom.setter
    def padding_bottom(self, p):
        wx, wy = self._padding[0]
        hx, hy = self._padding[1]
        self.padding = ((wx, wy), (p, hy))

    @property
    def spacings(self):
        if self.managed or self.graphics.inset:
            return [0, 0]
        else:
            return self._spacings
    @spacings.setter
    def spacings(self, spacings):
        if not (self.managed or self.graphics.inset):
            try:
                w, h = spacings
            except ValueError:
                w = h = spacings

            """
            The width of the padding between subplots, as a fraction of the average Axes width.
            """
            if hasattr(self.axes, 'get_bboxes'):
                bboxes = self.axes.get_bboxes()
            else:
                bboxes = self.figure.get_bboxes()
            # bboxes = [a.get_bbox() for a in self.axes]
            # W = bbox[1][0] - bbox[0][0]
            # H = bbox[1][1] - bbox[0][1]
            W = sum(b[1][0] - b[0][0] for b in bboxes) / len(bboxes)
            H = sum(b[1][1] - b[0][1] for b in bboxes) / len(bboxes)

            # W, H = self.image_size
            if w < 1:
                wp = round(w * W)
            else:
                wp = w
            if h < 1:
                hp = round(h * H)
            else:
                hp = h
            self._spacings = (wp, hp)

            w = wp / W
            h = hp / H

            self.figure.set_figure_spacings([w, h])


    @property
    def colorbar(self):
        return self._colorbar
    @colorbar.setter
    def colorbar(self, c):
        self._colorbar = c
        # if self._cbar_obj is not None:
        #     self.graphics.remove(self._cbar_obj)
        if self._cbar_obj is None:
            if self._colorbar is True:
                self._cbar_obj = self.graphics.add_colorbar()
            elif isinstance(self._colorbar, dict):
                self._cbar_obj = self.graphics.add_colorbar(**self.colorbar)
        elif self._colorbar is None:
            pass
            #self.graphics.remove(self._cbar_obj)

class GraphicsPropertyManager3D(GraphicsPropertyManager):
    def __init__(self, graphics, figure, axes, managed=False):
        super().__init__(graphics, figure, axes, managed=managed)
        self._view_settings = None

    @property
    def axes_labels(self):
        return self._axes_labels
    @axes_labels.setter
    def axes_labels(self, labels):
        if self._axes_labels is None:
            self._axes_labels = (self.axes.get_xlabel(), self.axes.get_ylabel(), self.axes.get_zlabel())
        try:
            xlab, ylab, zlab = labels
        except ValueError:
            xlab, ylab, zlab = labels = (labels, self._axes_labels[1], self._axes_labels[2])

        self._axes_labels = tuple(labels)
        if xlab is None:
            self.axes.set_xlabel("")
        elif isinstance(xlab, Styled):
            self.axes.set_xlabel(*xlab.val, **xlab.opts)
        elif Styled.could_be(xlab):
            xlab = Styled.construct(xlab)
            self.axes.set_xlabel(*xlab.val, **xlab.opts)
        else:
            self.axes.set_xlabel(xlab)

        if ylab is None:
            self.axes.set_ylabel("")
        elif isinstance(ylab, Styled):
            self.axes.set_ylabel(*ylab.val, **ylab.opts)
        elif Styled.could_be(ylab):
            ylab = Styled.construct(ylab)
            self.axes.set_ylabel(*ylab.val, **ylab.opts)
        else:
            self.axes.set_ylabel(ylab)

        if zlab is None:
            self.axes.set_zlabel("")
        elif isinstance(zlab, Styled):
            self.axes.set_zlabel(*zlab.val, **zlab.opts)
        elif Styled.could_be(zlab):
            zlab = Styled.construct(zlab)
            self.axes.set_zlabel(*zlab.val, **zlab.opts)
        else:
            self.axes.set_zlabel(zlab)

    @property
    def plot_range(self):
        if self._plot_range is None:
            pr = (self.axes.get_xlim(), self.axes.get_ylim())
        else:
            pr = self._plot_range
        return pr
    @plot_range.setter
    def plot_range(self, ranges):
        if self._plot_range is None:
            self._plot_range = (self.axes.get_xlim(), self.axes.get_ylim(), self.axes.get_zlim())
        elif len(self._plot_range) == 2:
            self._plot_range = self._plot_range + (self.axes.get_zlim(),)

        try:
            x, y, z = ranges
        except ValueError:
            try:
                x, y = ranges
            except ValueError:
                raise ValueError("bad plot range spec {}".format(ranges))
            else:
                if isinstance(x, (int, float)):
                    x, y, z = ranges = (self._plot_range[0], self._plot_range[1], ranges)
                else:

                    z = self._plot_range[2]
        # else:
        #     if isinstance(x, int) or isinstance(x, float):
        #         x, y, z = ranges = (self._plot_range[0], self._plot_range[1], ranges)

        self._plot_range = tuple(ranges)

        if isinstance(x, Styled):  # name feels wrong here...
            self.axes.set_xlim(*x.val, **x.opts)
        elif x is not None:
            self.axes.set_xlim(x)
        if isinstance(y, Styled):
            self.axes.set_ylim(*y.val, **y.opts)
        elif y is not None:
            self.axes.set_ylim(y)
        if isinstance(z, Styled):
            self.axes.set_zlim(*z.val, **z.opts)
        elif z is not None:
            self.axes.set_zlim(z)

    def _set_xticks(self, x, **opts):
        return self._set_ticks(x,
                               set_ticks=self.axes.set_xticks,
                               set_locator=self.axes.xaxis.set_major_locator,
                               set_minor_locator=self.axes.xaxis.set_minor_locator,
                               **opts
                               )
    def _set_yticks(self, y, **opts):
        return self._set_ticks(y,
                               set_ticks=self.axes.set_yticks,
                               set_locator=self.axes.yaxis.set_major_locator,
                               set_minor_locator=self.axes.yaxis.set_minor_locator,
                               **opts
                               )
    def _set_zticks(self, z, **opts):
        return self._set_ticks(z,
                               set_ticks=self.axes.set_zticks,
                               set_locator=self.axes.zaxis.set_major_locator,
                               set_minor_locator=self.axes.zaxis.set_minor_locator,
                               **opts
                               )

    @property
    def ticks(self):
        return self._ticks
    @ticks.setter
    def ticks(self, ticks):
        if self._ticks is None:
            self._ticks = (self.axes.get_xticks(), self.axes.get_yticks(), self.axes.get_zticks())
        if ticks is not None:
            try:
                x, y, z = ticks
            except ValueError:
                x, y, z = ticks = (self._ticks[0], self._ticks[1], ticks)

            self._ticks = ticks

            self._set_xticks(x)
            self._set_yticks(y)
            self._set_zticks(z)

    @property
    def ticks_style(self):
        return self._ticks_style
    @ticks_style.setter
    def ticks_style(self, ticks_style):
        if self._ticks_style is None:
            self._ticks_style = (None,)*3
        try:
            x, y, z = ticks_style
        except ValueError:
            x, y, z = ticks_style = (self._ticks_style[0], self._ticks_style[1], ticks_style)
        self._ticks_style = ticks_style
        if x is not None:
            self.axes.set_xtick_style(
                **x
            )
        if y is not None:
            self.axes.set_ytick_style(
                **y
            )
        if z is not None:
            self.axes.set_ztick_style(
                **z
            )

    @property
    def view_settings(self):
        return self.axes.get_view_settings()
    @view_settings.setter
    def view_settings(self, value):
        self.axes.set_view_settings(**value)


"""
Provides Graphics base classes that can be extended upon
"""

import os, weakref, numpy as np, functools

from .Properties import GraphicsPropertyManager, GraphicsPropertyManager3D
from .Styling import Styled, ThemeManager, PlotLegend
from .Backends import GraphicsBackend, DPI_SCALING

__all__ = ["GraphicsBase", "Graphics", "Graphics3D", "GraphicsGrid"]
__reload_hook__ = ['.Properties', ".Styling", ".Backends"]


class GraphicsException(Exception):
    pass

########################################################################################################################
#
#                                               GraphicsBase
#
from abc import *


class FigureTreeManager:
    _figure_mapping = weakref.WeakValueDictionary()
    _figure_children = weakref.WeakKeyDictionary()
    @classmethod
    def resolve_figure_graphics(cls, fig):
        if fig in cls._figure_mapping:
            return cls._figure_mapping[fig]
    @classmethod
    def add_figure_graphics(cls, fig, graphics):
        if fig in cls._figure_mapping:
            parent = cls._figure_mapping[fig]
            # cls._figure_children[parent].add(graphics)
            cls._figure_children[parent].append(graphics)
        else:
            cls._figure_mapping[fig] = graphics
            cls._figure_children[graphics] = []
    @classmethod
    def remove_figure_mapping(cls, fig):
        if fig in cls._figure_mapping:
            parent = cls._figure_mapping[fig]
            del cls._figure_mapping[fig]
            del cls._figure_children[parent]
    @classmethod
    def get_child_graphics(cls, fig):
        if fig in cls._figure_mapping:
            parent = cls._figure_mapping[fig]
            return cls._figure_children[parent]

    _axes_mapping = weakref.WeakValueDictionary()
    _axes_children = weakref.WeakKeyDictionary()
    @classmethod
    def resolve_axes_graphics(cls, axes):
        if hasattr(axes, 'figure'):  # ignore GraphicsGrids
            if axes in cls._axes_mapping:
                return cls._axes_mapping[axes]
    @classmethod
    def add_axes_graphics(cls, axes, graphics):
        if hasattr(axes, 'figure'):
            if axes in cls._axes_mapping:
                parent = cls._axes_mapping[axes]
                cls._axes_children[parent].append(graphics)
            else:
                cls._axes_mapping[axes] = graphics
                cls._axes_children[graphics] = []
    @classmethod
    def remove_axes_mapping(cls, axes):
        if hasattr(axes, 'figure'):
            if axes in cls._axes_mapping:
                parent = cls._axes_mapping[axes]
                del cls._axes_mapping[axes]
                del cls._axes_children[parent]
    @classmethod
    def get_axes_child_graphics(cls, axes):
        if hasattr(axes, 'figure'):
            if axes in cls._axes_mapping:
                parent = cls._axes_mapping[axes]
                return cls._axes_children[parent]


class GraphicsBase(metaclass=ABCMeta):
    """
    The base class for all things Graphics
    Defines the common parts of the interface
    """


    opt_keys = {
        'background',
        # 'axes_labels',
        # 'plot_label',
        # 'plot_range',
        # 'plot_legend',
        # 'legend_style',
        # 'ticks',
        # 'ticks_style',
        # 'ticks_label_style',
        # 'axes_bbox',
        # 'scale',
        'padding',
        'spacings',
        # 'aspect_ratio',
        'image_size',
        'event_handlers',
        'animated',
        'epilog'
        'prolog',
        'subplot_kw',
        'theme',
        'annotations'
    }

    @staticmethod
    def _split_props_list(props, filter_set):
        excl = {k: props[k] for k in props.keys() - filter_set}
        incl = {k: props[k] for k in props.keys() & filter_set}
        return incl, excl

    def get_raw_attr(self, key):
        exc = None
        try:
            v = object.__getattribute__(self, '_' + key)  # we overloaded getattr
        except AttributeError as e:
            try:
                v = object.__getattribute__(self._prop_manager, '_' + key)  # we overloaded getattr
            except AttributeError as e:
                exc = e
            else:
                exc = None
        if exc is not None:
            raise exc # from None
        return v

    def _get_def_opt(self, key, val, theme, parent=None):
        if val is None:
            try:
                v = object.__getattribute__(self, '_'+key) # we overloaded getattr
            except AttributeError as e:
                try:
                    v = object.__getattribute__(self._prop_manager, '_' + key)  # we overloaded getattr
                except AttributeError:
                    if parent is not None:
                        try:
                            v = parent.get_raw_attr(key)
                        except AttributeError:
                            v = None
                    else:
                        v = None
            if v is None and key in self.default_style:
                v = self.default_style[key]
            if v is None and theme is not None and key in theme:
                v = theme[key]
            return v
        else:
            return val

    def _update_copy_opt(self, key, val):
        if not self._in_init:
            has_val = False
            if key not in self._init_opts or self._init_opts[key] is None:
                has_val = True
                old = self._get_def_opt(key, None, self.theme)
            elif key in self._init_opts:
                has_val = True
                old = self._init_opts[key]
            else:
                old = None
            if has_val:
                changed = val != old
                if isinstance(changed, np.ndarray):
                    changed = changed.any()
                if changed:
                    self._init_opts[key] = val

    layout_keys = {
        'name',
        'figure',
        'tighten',
        'axes',
        'subplot_kw',
        'parent',
        'image_size',
        'figsize',
        'padding',
        'aspect_ratio',
        'interactive',
        'reshowable',
        'theme',
        'prop_manager',
        'theme_manager',
        'managed',
        # 'inset',
        'event_handlers',
        'animated',
        'prolog',
        'epilog',
        'annotations',
        'backend',
        'backend_options'
    }
    def __init__(self,
                 *args,
                 name=None,
                 figure=None,
                 tighten=False,
                 axes=None,
                 subplot_kw=None,
                 parent=None,
                 image_size=None,
                 padding=None,
                 aspect_ratio=None,
                 interactive=None,
                 reshowable=None,
                 backend='matplotlib',
                 backend_options=None,
                 theme=None,
                 prop_manager=GraphicsPropertyManager,
                 theme_manager=ThemeManager,
                 managed=None,
                 # inset=False,
                 strict=True,
                 # annotations=None,
                 **opts
                 ):
        """
        :param args:
        :type args:
        :param figure:
        :type figure: GraphicsBackend.Figure | None
        :param axes:
        :type axes: GraphicsBackend.Figure.Axes | None
        :param subplot_kw:
        :type subplot_kw: dict | None
        :param parent:
        :type parent: GraphicsBase | None
        :param opts:
        :type opts:
        """

        self.name = name

        self._init_opts = dict(opts, # for copying
                               tighten=tighten,
                               axes=axes,
                               subplot_kw=subplot_kw,
                               image_size=image_size,
                               padding=padding,
                               aspect_ratio=aspect_ratio,
                               interactive=interactive,
                               reshowable=reshowable,
                               backend=backend,
                               backend_options=backend_options,
                               theme=theme,
                               prop_manager=prop_manager,
                               theme_manager=theme_manager,
                               managed=managed
                               )
        self._in_init = True

        if subplot_kw is None:
            subplot_kw = {}

        self.interactive = interactive
        self.backend = GraphicsBackend.lookup(backend, backend_options)
        if isinstance(figure, GraphicsBase): parent = figure
        inherit_layout = parent is not None and (axes is None or parent.axes is axes) # check inset
        if inherit_layout: prop_manager = parent._prop_manager

        theme = self._get_def_opt('theme', theme, {})
        self.theme=theme
        self.theme_manager=theme_manager
        # if theme is not None:
        #     theme_dict = theme_manager.resolve_theme(theme)[1]
        # else:
        #     theme_dict = {}

        if managed is None:
            if figure is not None and isinstance(figure, GraphicsBase):
                managed = figure.managed
            else:
                managed = False
        self.managed = managed
        self._inset = None

        interactive = self._get_def_opt('interactive', interactive, theme)
        self.interactive = interactive

        reshowable = self._get_def_opt('reshowable', reshowable, theme)
        self.reshowable = reshowable

        theme_parent = parent if inherit_layout else None
        aspect_ratio = self._get_def_opt('aspect_ratio', aspect_ratio,  theme, theme_parent)
        image_size = self._get_def_opt('image_size', image_size, theme, theme_parent)
        padding = self._get_def_opt('padding', padding, theme, theme_parent)
        if figure is None and image_size is not None and 'figsize' not in subplot_kw:
            try:
                w, h = image_size
            except TypeError:
                w = image_size
                asp = aspect_ratio
                if asp is None or isinstance(asp, str):
                    asp = 4.8/6.4
                h = w * asp
            if padding is not None:
                pw, ph = padding
                try:
                    pwx, pwy = pw
                except (TypeError, ValueError):
                    pwx = pwy = pw
                try:
                    phx, phy = ph
                except (TypeError, ValueError):
                    phx = phy = ph
                w += pwx + pwy
                h += phx + phy
                image_size = (w, h)
            subplot_kw['figsize'] = (w/DPI_SCALING, h/DPI_SCALING)

        self.subplot_kw = subplot_kw
        with self.theme_manager.from_spec(self.theme, backend=self.backend):
            self.figure, self.axes = self.initialize_figure_and_axes(figure, axes, *args, **subplot_kw)
        self.figure._called_show = False # for avoiding excess show calls with a custom backend
        if self.inset:
            FigureTreeManager.add_axes_graphics(self.axes, self)
        else:
            FigureTreeManager.add_figure_graphics(self.figure, self)

        if not self.interactive: # TODO: should this really be done here...?
            self.backend.disable_interactivity()

        if isinstance(prop_manager, type):
            prop_manager = prop_manager(self, self.figure, self.axes, managed=managed)
        self._prop_manager = prop_manager
        self._colorbar_axis = None
        self.set_options(padding=padding, aspect_ratio=aspect_ratio, image_size=image_size, strict=strict, **opts)

        self.event_handler = None
        self._shown = False
        self.animator = None
        self.tighten = tighten

        self._in_init = False

    def initialize_figure_and_axes(self, figure, axes, *args, **kw) -> 'tuple[GraphicsBackend.Figure, GraphicsBackend.Figure.Axes]':
        """Initializes the subplots for the Graphics object

        :param figure:
        :type figure:
        :param axes:
        :type axes:
        :param args:
        :type args:
        :param kw:
        :type kw:
        :return: figure, axes
        :rtype: GraphicsFigure, GraphicsAxes
        """

        if figure is None:
            figure, axes = self.backend.create_figure(*args, **kw)
        elif isinstance(figure, GraphicsBase) or all(hasattr(figure, h) for h in ['layout_keys']):
            if axes is None:
                axes = figure.axes # type: GraphicsAxes
            figure = figure.figure # type: GraphicsFigure

        self.figure = figure
        self.axes = axes # we set these temporarily to make inset work right
        if self.inset:
            FigureTreeManager.add_axes_graphics(self.axes, self)
        else:
            FigureTreeManager.add_figure_graphics(self.figure, self)

        if axes is None:
            raise NotImplementedError("dealing with just getting a figure object not implemented...")
            # if isinstance(figure, GraphicsBase):
            #
            # if not hasattr(figure, 'create_axes'):
            #     figure = figure.figure
            # if not hasattr(figure, 'axes'):
            #     axes = figure.create_axes(1, 1, 1) # type: GraphicsAxes
            # else:
            #     axes = figure.axes

        return figure, axes

    @property
    def parent(self):
        if self.inset:
            return FigureTreeManager.resolve_axes_graphics(self.axes)
        else:
            return FigureTreeManager.resolve_figure_graphics(self.figure)
    @property
    def figure_parent(self):
        return FigureTreeManager.resolve_figure_graphics(self.figure)
    @property
    def inset(self):
        fp = self.figure_parent
        return not self.managed and fp is not None and (self.axes is not fp.axes)
    @property
    def children(self):
        if self.parent is self:
            if self.inset:
                return FigureTreeManager.get_axes_child_graphics(self.axes)
            else:
                return FigureTreeManager.get_child_graphics(self.figure)
        else:
            return None

    @property
    def event_handlers(self):
        from .Interactive import EventHandler
        h = self.event_handler  # type: EventHandler
        if h is not None:
            h = h.data
        return h

    @property
    def animated(self):
        return self._animated

    def bind_events(self, *handlers, **events):
        from .Interactive import EventHandler

        if len(handlers) > 0 and isinstance(handlers[0], dict):
            handlers = handlers[0]
        elif len(handlers) == 0 or (len(handlers) > 0 and handlers[0] is not None):
            handlers = dict(handlers)
        if isinstance(handlers, dict):
            handlers = dict(handlers, **events)
            if self.event_handler is None:
                self.event_handler = EventHandler(self, **handlers)
            else:
                self.event_handler.bind(**handlers)

    def create_animation(self, *args, **opts):
        from .Interactive import Animator

        if len(args) > 0 and args[0] is not None:
            if self.animator is not None:
                self.animator.stop()
            self.animator = Animator(self, *args, **opts)

    def animate_frames(self, frames, **opts):
        return self.figure.animate_frames(frames, **opts)

    known_keys = layout_keys
    def _check_opts(self, opts):
        diff = opts.keys() - self.known_keys
        if len(diff) > 0:
            raise ValueError("unknown options for {}: {}".format(
                type(self).__name__, list(diff)
            ))
    def set_options(self,
                    event_handlers=None,
                    animated=None,
                    prolog=None,
                    epilog=None,
                    strict=True,
                    **opts
                    ):
        """Sets options for the plot
        :param event_handlers:
        :param animated:
        :param opts:
        :type opts:
        :return:
        :rtype:
        """
        if strict:
            self._check_opts(opts)

        self.bind_events(event_handlers)
        self._animated = animated
        self.create_animation(animated)

        if prolog is not None or not hasattr(self, '_prolog'):
            self._prolog = prolog
            if self._prolog is not None:
                self.prolog = prolog

        if epilog is not None or not hasattr(self, '_epilog'):
            self._epilog = epilog
            if self._epilog is not None:
                self.epilog = epilog

    @property
    def prolog(self):
        return self._prolog
    @prolog.setter
    def prolog(self, p):
        self._update_copy_opt('prolog', p)
        # might want to clear the elements in the prolog?
        self._prolog = p

    @property
    def epilog(self):
        return self._epilog
    @epilog.setter
    def epilog(self, e):
        self._update_copy_opt('epilog', e)
        # might want to clear the elements in the epilog?
        self._epilog = e

    @property
    def opts(self):
        opt_dict = {}
        for k in self.opt_keys:
            if (
                    k in self.__dict__
                    or hasattr(self._prop_manager, k)
            ):
                opt_dict[k] = getattr(self, k)
            elif (
                    "_"+k in self.__dict__
                    or hasattr(self, "_"+k)
            ):
                opt_dict[k] = getattr(self, "_" + k)
        return opt_dict

    def copy(self, **kwargs):
        """Creates a copy of the object with new axes and a new figure

        :return:
        :rtype:
        """
        return self.change_figure(None, **kwargs)
    def _get_init_opts(self, parent_opts, unmerged_keys=None):
        if unmerged_keys is None:
            unmerged_keys = self.layout_keys
        base = self._init_opts.copy()
        problems = unmerged_keys & base.keys()
        for k in problems & parent_opts.keys():
            del base[k]
        return base
    def change_figure(self, new, *init_args, figs=None, **init_kwargs):
        """Creates a copy of the object with new axes and a new figure

        :return:
        :rtype:
        """
        parent = self.figure_parent
        figs = {} if figs is None else figs
        if parent is self:
            if new.figure is self.figure:
                figs[self] = self
                for c in self.children:
                    figs[c] = c
            else:
                # opts = self.get_init_opts(**init_kwargs)
                if self not in figs:
                    figs[self] = self._change_figure(new, *init_args, parent_opts={}, **init_kwargs)
                base = figs[self]
                parent_opts = dict(self._get_init_opts({}), **init_kwargs)
                for c in self.children:
                    if c not in figs:
                        figs[c] = c._change_figure(base, parent_opts=parent_opts)
        else:
            parent.change_figure(new, *init_args, figs=figs, **init_kwargs)
        return figs[self]
    def _get_init_args(self, *init_args):
        return init_args
    def _change_figure(self, new, *init_args, parent_opts=None, **init_kwargs):
        """Creates a copy of the object with new axes and a new figure

        :return:
        :rtype:
        """
        return type(self)(
            *self._get_init_args(*init_args),
            **dict(
                self._get_init_opts(parent_opts),
                figure=new,
                **init_kwargs
            )
        )

    def _prep_show(self, parent=False):
        self.set_options(**self.opts)  # matplotlib is dumb so it makes sense to just reset these again...
        if self.prolog is not None:
            self._prolog_graphics = [p.plot(self.axes, graphics=self) for p in self.prolog] # not sure this is doing what it should...
        if self.epilog is not None:
            self._epilog_graphics = [e.plot(self.axes, graphics=self) for e in self.epilog]
        if self.tighten:
            self.figure.tight_layout()
        self._shown = True
        return self

    def prep_show(self):
        if self.figure_parent is self:
            self._prep_show(parent=True)
            for c in self.children:
                if hasattr(c, '_prep_show'):
                    c._prep_show(parent=False)
                else:
                    c.prep_show()
        else:
            self.figure_parent.prep_show()
        return self
    def show(self, reshow=None):
        if reshow or not self._shown:
            self.prep_show()
            if not self.managed:
                ni = not self.interactive
                try:
                    if ni:
                        self.backend.enable_interactivity()
                    self.backend.show_figure(self.figure, reshow=reshow)
                finally:
                    self.interactive = not ni
                    if ni:
                        self.backend.disable_interactivity()
        else:
            if self.reshowable:
                return self.show(reshow=True)
            else:
                # self._shown = False
                return self.copy().show()
                # raise GraphicsException("{}.show can only be called once per object".format(type(self).__name__))

        return self.backend.show_figure(self.figure, reshow=reshow)

    def close(self, force=False):
        if (
                force
                or self.figure not in FigureTreeManager._figure_mapping
                or FigureTreeManager.resolve_figure_graphics(self.figure) is self
                or FigureTreeManager.resolve_axes_graphics(self.axes) is self
        ): # parent manages cleanup
            if self.inset:
                # print("removing inset axes: {}".format(self))
                self.backend.remove_axes(self.axes)
            else:
                # print("closing: {}".format(self))
                self.backend.close_figure(self.figure)
                FigureTreeManager.remove_figure_mapping(self.figure)
        # else:
        #     print("close failed: {}".format(self))

    def __del__(self):
        try:
            self.close()
        except AttributeError:
            pass

    def __repr__(self):
        return "{}({}, figure={}<{}>)".format(
            type(self).__name__,
            id(self) if self.name is None else self.name,
            self.figure,
            id(self.figure)
        )

    def clear(self):
        self.backend.clear_axes(self.axes)
        # FigureTreeManager.remove_figure_mapping(self.figure)

    _display_locks = set()
    def _ipython_display_(self):
        if self not in self._display_locks:  # don't want to call this over and over...
            self._display_locks.add(self)
            try:
                self.show()
            finally:
                self._display_locks.remove(self)
    def _repr_html_(self):
        # hacky, but hopefully enough to make it work?
        return self.figure._repr_html_()

    def savefig(self, where, format=None, **kw):
        """
        Saves the image to file
        :param where:
        :type where:
        :param format:
        :type format:
        :param kw:
        :type kw:
        :return: file it was saved to (I think...?)
        :rtype: str
        """
        if 'facecolor' not in kw:
            kw['facecolor'] = self.background
        if format is None:
            format = os.path.splitext(where)[1].split('.')[-1]
        self.prep_show()
        return self.figure.savefig(where,
                    format=format,
                    **kw
                    )
    def to_png(self):
        """
        Used by Jupyter and friends to make a version of the image that they can display, hence the extra 'tight_layout' call
        :return:
        :rtype:
        """
        import io
        buf = io.BytesIO()
        self.prep_show()
        self.figure.tight_layout()
        fig = self.figure
        fig.savefig(buf,
                    format='png',
                    facecolor=self.background #-_- stupid MPL
                    )
        buf.seek(0)
        return buf

    def _repr_png_(self):
        return self.to_png().read()

    def create_colorbar_axis(self,
                             figure=None,
                             size=(20, 200),
                             tick_padding=40
                             ):
        fig = self.figure if figure is None else figure
        # if self._colorbar_axis is None:
        # TODO: I'd like to have better control over how much space this colorbar takes in the future
        #   it might be a mistake to make this not percentage based...
        W, H = self.image_size
        if size[0] < 1:
            size = (W * size[0], H * size[1])
        size = (size[0] + tick_padding, size[1])
        cur_padding = self.padding
        if cur_padding[0][1] < size[0] + 5:
            self.padding_right = size[0] + 5
            new_padding = self.padding
            wpad_old = cur_padding[0][0] + cur_padding[0][1]
            wpad_new = new_padding[0][0] + new_padding[0][1]
            wdiff = wpad_new - wpad_old
        else:
            wdiff = 0
        # we need to now shrink the spacing by enough to compensate for this
        # it would be best to mess with the figure sizes themselves, but this is the easier
        # solution for the moment...
        sp = self.spacings
        if sp is not None:
            ws, hs = sp
            nspaces = self.shape[1] - 1
            ws = ws - (wdiff / nspaces)
            ws = max(ws, 0)
            self.spacings = (ws, hs)
        cbw = size[0] / W
        tw = tick_padding / W
        cbh = size[1] / H
        xpos = 1 - cbw
        ypos = .5 - cbh / 2  # new_padding[1][0]/H

        bbox = [[xpos, ypos], [xpos + cbw - tw, ypos + cbh]]
        theme = self.theme
        if self.theme is not None:
            with self.theme_manager.from_spec(theme, backend=self.backend):
                axis = fig.create_inset(bbox)
        else:
            axis = fig.create_inset(bbox)
        return axis
    def add_colorbar(self,
                     graphics=None,
                     norm=None,
                     cmap=None,
                     size=(20, 200),
                     tick_padding=40,
                     colorbar_axes=None,
                     cax=None,
                     **kw
                     ):
        fig = self.figure  # type: GraphicsBackend.Figure
        ax = self.axes  # type: GraphicsBackend.Figure.Axes

        if cax is not None:
            graphics = cax
        if colorbar_axes is None:
            if self._colorbar_axis is None:
                colorbar_axes = self.create_colorbar_axis(
                    size=(20, 200),
                    tick_padding=40,
                )
        if self._colorbar_axis is None:
            self._colorbar_axis = colorbar_axes
        if colorbar_axes is not self._colorbar_axis:
            self._colorbar_axis.remove()
            self._colorbar_axis = colorbar_axes

        return fig.create_colorbar(graphics, colorbar_axes,
                                   norm=norm,
                                   cmap=cmap,
                                   **kw)

    axes_params = {"adjustable", "agg_filter", "alpha", "anchor", "animated", "aspect",
                   "autoscale_on", "autoscalex_on", "autoscaley_on", "axes_locator", "axisbelow",
                   "box_aspect", "clip_box", "clip_on", "clip_path", "facecolor",
                   "frame_on", "gid", "in_layout", "label", "navigate",
                   "navigate_mode", "path_effects", "picker", "position",
                   "prop_cycle", "rasterization_zorder", "rasterized", "sketch_params",
                   "snap", "title", "transform", "url", "visible", "xbound",
                   "xlabel", "xlim", "xmargin", "xscale", "xticklabels", "xticks",
                   "ybound", "ylabel", "ylim", "ymargin", "yscale", "yticklabels",
                   "yticks", "zorder"} # not sure what to do with these yet...
    inset_options = dict()
    axes_keys = set()
    _axes_padding_offset = [1, 0]
    def create_inset(self, bbox, coordinates='scaled', graphics_class=None, **opts):
        if hasattr(bbox, 'get_points'):
            bbox = bbox.get_points()
        ((lx, by), (rx, ty)) = bbox

        if coordinates == 'absolute':
            raise NotImplementedError("can't construct inset axes with absolute coordinates")
        elif coordinates == 'scaled': # scaled to live within the frame
            ((alx, aby), (arx, aty)) = self.axes.bbox.get_points()
            alx = alx + self._axes_padding_offset[0]
            arx = arx + self._axes_padding_offset[0]
            aby = aby + self._axes_padding_offset[1]
            aty = aty + self._axes_padding_offset[1]
            w, h = self.figure_parent.image_size
            # we scale the coordinates to live within the axes screenbox
            sl = alx / w
            sr = arx / w
            sw = (sr - sl)
            sb = aby / h
            st = aty / h
            sh = (st - sb)
            lx = lx * sw + sl
            rx = rx * sw + sl
            by = by * sh + sb
            ty = ty * sh + sb

        if graphics_class is None:
            graphics_class = type(self)
        opts = dict(self.inset_options, **opts)
        # raise Exception(opts, self.axes_keys)
        # ax_par, fig_par = self._split_props_list(opts, self.axes_keys)

        with self.theme_manager.from_spec(self.theme, backend=self.backend):
            ax = self.figure.add_axes([lx, by, rx-lx, ty-by])
        return graphics_class(figure=self, axes=ax, **opts)

########################################################################################################################
#
#                                               Graphics
#
class Graphics(GraphicsBase):

    default_style = dict(
        theme='mccoy',
        frame=((True, False), (True, False)),
        image_size=(370, 345),
        padding=((60, 10), (35, 10)),
        interactive=False,
        reshowable=False,
        aspect_ratio='auto'
    )

    axes_keys = {
        'plot_label',
        'plot_legend',
        'legend_style'
        'axes_labels',
        'frame',
        'frame_style',
        'plot_range',
        'ticks',
        'ticks_style',
        'ticks_label_style',
        'axes_bbox'
    }
    figure_keys = {
        'scale',
        'aspect_ratio'
        'image_size',
        'padding',
        'spacings',
        'background',
        'colorbar'
    }
    layout_keys = axes_keys | figure_keys | GraphicsBase.layout_keys
    known_keys = layout_keys
    def set_options(self,
                    axes_labels=None,
                    plot_label=None,
                    plot_range=None,
                    plot_legend=None,
                    legend_style=None,
                    frame=None,
                    frame_style=None,
                    ticks=None,
                    scale=None,
                    padding=None,
                    spacings=None,
                    ticks_style=None,
                    ticks_label_style=None,
                    image_size=None,
                    axes_bbox=None,
                    aspect_ratio=None,
                    background=None,
                    colorbar=None,
                    prolog=None,
                    epilog=None,

                    **parent_opts
                    ):

        super().set_options(prolog=prolog, epilog=epilog, **parent_opts)

        opts = (
            ('plot_label', plot_label),
            ('plot_legend', plot_legend),
            ('legend_style', legend_style),
            ('axes_labels', axes_labels),
            ('frame', frame),
            ('frame_style', frame_style),
            ('plot_range', plot_range),
            ('ticks', ticks),
            ('ticks_style', ticks_style),
            ('ticks_label_style', ticks_label_style),
            ('scale', scale),
            ('aspect_ratio', aspect_ratio),
            ('image_size', image_size),
            ('axes_bbox', axes_bbox),
            ('padding', padding),
            ('spacings', spacings),
            ('background', background),
            ('colorbar', colorbar)
        )
        for oname, oval in opts:
            oval = self._get_def_opt(oname, oval, {})
            if oval is not None:
                setattr(self, oname, oval)

    @property
    def artists(self):
        return []

    # attaching custom property setters
    @property
    def plot_label(self):
        return self._prop_manager.plot_label
    @plot_label.setter
    def plot_label(self, value):
        self._update_copy_opt('plot_label', value)
        self._prop_manager.plot_label = value

    @property
    def plot_legend(self):
        return self._prop_manager.plot_legend
    @plot_legend.setter
    def plot_legend(self, value):
        self._update_copy_opt('plot_legend', value)
        self._prop_manager.plot_legend = value

    @property
    def legend_style(self):
        return self._prop_manager.legend_style
    @legend_style.setter
    def legend_style(self, value):
        self._update_copy_opt('legend_style', value)
        self._prop_manager.legend_style = value

    @property
    def axes_labels(self):
        return self._prop_manager.axes_labels
    @axes_labels.setter
    def axes_labels(self, value):
        self._update_copy_opt('axes_labels', value)
        self._prop_manager.axes_labels = value

    @property
    def frame(self):
        return self._prop_manager.frame
    @frame.setter
    def frame(self, value):
        self._update_copy_opt('frame', value)
        self._prop_manager.frame = value

    @property
    def frame_style(self):
        return self._prop_manager.frame_style
    @frame_style.setter
    def frame_style(self, value):
        self._update_copy_opt('frame_style', value)
        self._prop_manager.frame_style = value

    @property
    def plot_range(self):
        return self._prop_manager.plot_range
    @plot_range.setter
    def plot_range(self, value):
        self._update_copy_opt('plot_range', value)
        self._prop_manager.plot_range = value

    @property
    def ticks(self):
        return self._prop_manager.ticks
    @ticks.setter
    def ticks(self, value):
        self._update_copy_opt('ticks', value)
        self._prop_manager.ticks = value

    @property
    def ticks_style(self):
        return self._prop_manager.ticks_style
    @ticks_style.setter
    def ticks_style(self, value):
        self._update_copy_opt('ticks_style', value)
        self._prop_manager.ticks_style = value

    @property
    def ticks_label_style(self):
        return self._prop_manager.ticks_label_style
    @ticks_label_style.setter
    def ticks_label_style(self, value):
        self._update_copy_opt('ticks_label_style', value)
        self._prop_manager.ticks_label_style = value

    @property
    def scale(self):
        return self._prop_manager.scale
    @scale.setter
    def scale(self, value):
        self._update_copy_opt('scale', value)
        self._prop_manager.scale = value

    @property
    def axes_bbox(self):
        return self._prop_manager.axes_bbox
    @axes_bbox.setter
    def axes_bbox(self, value):
        self._update_copy_opt('axes_bbox', value)
        self._prop_manager.axes_bbox = value

    @property
    def aspect_ratio(self):
        return self._prop_manager.aspect_ratio
    @aspect_ratio.setter
    def aspect_ratio(self, value):
        self._update_copy_opt('aspect_ratio', value)
        self._prop_manager.aspect_ratio = value

    @property
    def image_size(self):
        return self._prop_manager.image_size
    @image_size.setter
    def image_size(self, value):
        self._update_copy_opt('image_size', value)
        self._prop_manager.image_size = value

    @property
    def figure_label(self):
        return self._prop_manager.figure_label
    @figure_label.setter
    def figure_label(self, value):
        self._update_copy_opt('figure_label', value)
        self._prop_manager.figure_label = value

    @property
    def padding(self):
        return self._prop_manager.padding
    @padding.setter
    def padding(self, value):
        self._update_copy_opt('padding', value)
        self._prop_manager.padding = value
    @property
    def padding_left(self):
        return self._prop_manager.padding_left
    @padding_left.setter
    def padding_left(self, value):
        self._update_copy_opt('padding_left', value)
        self._prop_manager.padding_left = value
    @property
    def padding_right(self):
        return self._prop_manager.padding_right
    @padding_right.setter
    def padding_right(self, value):
        self._update_copy_opt('padding_right', value)
        self._prop_manager.padding_right = value
    @property
    def padding_top(self):
        return self._prop_manager.padding_top
    @padding_top.setter
    def padding_top(self, value):
        self._update_copy_opt('padding_top', value)
        self._prop_manager.padding_top = value
    @property
    def padding_bottom(self):
        return self._prop_manager.padding_bottom
    @padding_bottom.setter
    def padding_bottom(self, value):
        self._update_copy_opt('padding_bottom', value)
        self._prop_manager.padding_bottom = value

    @property
    def spacings(self):
        return self._prop_manager.spacings
    @spacings.setter
    def spacings(self, value):
        self._update_copy_opt('spacings', value)
        self._prop_manager.spacings = value

    @property
    def background(self):
        return self._prop_manager.background
    @background.setter
    def background(self, value):
        self._update_copy_opt('background', value)
        self._prop_manager.background = value

    @property
    def colorbar(self):
        return self._prop_manager.colorbar
    @colorbar.setter
    def colorbar(self, value):
        self._update_copy_opt('colorbar', value)
        self._prop_manager.colorbar = value

    def _prep_show(self, parent=False):
        super()._prep_show()
        if parent:
            if self.plot_legend or any(hasattr(c, 'plot_legend') and c.plot_legend for c in self.children):
                pls = self.plot_legend
                ls = self.legend_style if self.legend_style is not None else {}
                PlotLegend.check_styles(ls)
                if isinstance(pls, PlotLegend):
                    self.axes.legend(handles=pls, **pls.opts, **ls)
                else:
                    self.axes.legend(**ls)
            if 'ticks' in self._init_opts:
                self.ticks = self._init_opts['ticks']

    def get_padding_offsets(self):
        ((l, r), (b, t)) = self.plot_range
        # w, h = self.image_size
        ((pl, pr), (pb, pt)) = self.padding
        ((alx, aby), (arx, aty)) = self.axes.get_bbox()
        w = arx - alx
        h = aty - aby
        pix_rat_x = (r - l) / w  # pixel to image coordiantes
        pix_rat_y = (t - b) / h
        ofl = pix_rat_x * (pl + self._axes_padding_offset[0])
        ofr = pix_rat_x * pr
        ofb = pix_rat_y * (pb + self._axes_padding_offset[1])
        oft = pix_rat_y * pt
        return [(ofl, ofr), (ofb, oft)]
    def get_bbox(self):
        # gives _effective_ image coordinates for the total space taken up by the figure
        plr = self.plot_range
        (ofl, ofr), (ofb, oft) = self.get_padding_offsets()
        bbox = [
            (plr[0][0] - ofl, plr[1][0] - ofb),
            (plr[0][1] + ofr, plr[1][1] + oft)
        ]
        return bbox
    inset_options = dict(
        image_size='auto',
        background='#FFFFFF00',
        frame_style={'linewidth':0},#, 'color':'#FFFFFF00'},
        aspect_ratio='auto',
        ticks=[[], []],
        ticks_style={'width':0},
        ticks_label_style=[{'fontsize':0}, {'fontsize':0}]
    )
    def create_inset(self, bbox, coordinates='absolute', graphics_class=None, **opts):
        if coordinates == 'absolute':
            ((lx, rx), (by, ty)) = self.plot_range
            w = rx - lx
            h = ty - by
            ((blx, bby), (brx, bty)) = bbox
            slx = ((blx - lx)) / w
            srx = ((brx - lx)) / w
            sby = ((bby - by)) / h
            sty = ((bty - by)) / h
            bbox = [(slx, sby), (srx, sty)]

            # ((ofl, ofr), (ofb, oft)) = self.get_padding_offsets()
            # w = (rx - lx + ofr + ofl)
            # h = (ty - by + oft + ofb)
            # # get coordinates relative to plot frame
            # ((blx, bby), (brx, bty)) = bbox
            # slx = (ofl+(blx-lx))/w
            # srx = (ofl+(brx-lx))/w
            # sby = (ofb+(bby-by))/h
            # sty = (ofb+(bty-by))/h
            # bbox = [(slx, sby), (srx, sty)]
            # # now shift to position of axis in total image
            coordinates = 'scaled'
        if graphics_class is None:
            graphics_class = Graphics
        return super().create_inset(bbox, coordinates=coordinates, graphics_class=graphics_class, **opts)

########################################################################################################################
#
#                                               Graphics3D
#
class Graphics3D(Graphics):

    opt_keys = GraphicsBase.opt_keys | {'view_settings', 'box_ratios'}
    known_keys = Graphics.opt_keys | {'animate'}

    # layout_keys = axes_keys | figure_keys | GraphicsBase.layout_keys
    # known_keys = layout_keys

    def __init__(self, *args,
                 figure=None,
                 axes=None,
                 subplot_kw=None,
                 event_handlers=None,
                 animate=None,
                 axes_labels=None,
                 plot_label=None,
                 plot_range=None,
                 plot_legend=None,
                 ticks=None,
                 scale=None,
                 ticks_style=None,
                 image_size=None,
                 background=None,
                 view_settings=None,
                 box_ratios=None,
                 backend='matplotlib3D',
                 **kwargs
                 ):

        self._backend = backend
        super().__init__(
            *args,
            figure=figure,
            axes=axes,
            subplot_kw=subplot_kw,
            axes_labels=axes_labels,
            plot_label=plot_label,
            plot_range=plot_range,
            plot_legend=plot_legend,
            ticks=ticks,
            scale=scale,
            ticks_style=ticks_style,
            image_size=image_size,
            event_handlers=event_handlers,
            animate=animate,
            view_settings=view_settings,
            backend=backend,
            background=background,
            box_ratios=box_ratios,
            prop_manager=GraphicsPropertyManager3D,
            **kwargs
        )

    def set_options(self,
                    view_settings=None,
                    box_ratios=None,
                    **parent_opts
                    ):

        super().set_options(**parent_opts)

        opts = (
            ('view_settings', view_settings),
            ('box_ratios', box_ratios),
        )
        for oname, oval in opts:
            oval = self._get_def_opt(oname, oval, {})
            if oval is not None:
                setattr(self, oname, oval)

    @property
    def box_ratios(self):
        return self._prop_manager.box_ratios
    @box_ratios.setter
    def box_ratios(self, value):
        self._prop_manager.box_ratios = value

    @property
    def view_settings(self):
        return self._prop_manager.view_settings
    @view_settings.setter
    def view_settings(self, value):
        self._prop_manager.view_settings = value

########################################################################################################################
#
#                                               GraphicsGrid
#
class GraphicsGrid(GraphicsBase):
    """
    A class for easily building sophisticated multi-panel figures.
    Robustification work still needs to be done, but the core interface is there.
    Supports themes & direct, easy access to the panels, among other things.
    Builds off of `GraphicsBase`.
    """
    default_style = dict(
        theme='mccoy',
        spacings=(50, 0),
        padding=((50, 10), (50, 10))
    )
    layout_keys = GraphicsBase.layout_keys | {'nrows', 'ncols'}
    known_keys = GraphicsBase.known_keys | {'graphics_class'}
    def __init__(self,
                 *args,
                 nrows=None, ncols=None,
                 graphics_class=Graphics,
                 figure=None,
                 axes=None,
                 subplot_kw=None,
                 subimage_size=(310, 310),
                 subimage_aspect_ratio='auto',
                 padding=None,
                 spacings=None,
                 **opts
                 ):

        if len(args) > 0:
            if len(args) > 1:
                raise ValueError("wat")
            args = args[0]
            if isinstance(args[0], GraphicsBase):
                nr = 1
                nc = len(args)
            else:
                nr = len(args)
                nc = max(len(a) for a in args)

            if nrows is None:
                nrows = nr
            if ncols is None:
                ncols = nc
        if nrows is None:
            nrows = 2
        if ncols is None:
            ncols = 2
        self.shape = (nrows, ncols)

        if subplot_kw is None:
            subplot_kw={}
        else:
            subplot_kw={'subplot_kw':subplot_kw}
        subplot_kw.update(dict(
            nrows=nrows, ncols=ncols,
            graphics_class=graphics_class,
            subimage_size=subimage_size,
            subimage_aspect_ratio=subimage_aspect_ratio,
            padding=padding,
            spacings=spacings
        ))
        super().__init__(
            figure=figure, axes=axes,
            graphics_class=graphics_class,
            subplot_kw=subplot_kw,
            padding=padding,
            spacings=spacings,
            **opts
        )
        self._colorbar_axis = None # necessary hack for only GraphicsGrid

        if len(args) > 0:
            if isinstance(args[0], GraphicsBase):
                for i, g in enumerate(args):
                    self[0, i] = g
            else:
                for i, r in enumerate(args):
                    for j, g in enumerate(r):
                        self[i, j] = g

    class GraphicsStack:
        def __init__(self, parent, graphics):
            self.parent = parent #type: GraphicsBase
            self.stack = np.empty_like(graphics, dtype=object)
            self.stack[:] = graphics
        def __getitem__(self, item):
            return self.stack[item]
        def __setitem__(self, item, value):
            self.stack[item] = value
        def _call_iter(self, attr):
            @functools.wraps(getattr(Graphics, attr))
            def call(*args, **kwargs):
                for ax in self.stack.flat:
                    yield getattr(ax, attr)(*args, **kwargs)
            return call
        def _axes_call_iter(self, attr):
            ax = next(self.stack.flat).axes
            @functools.wraps(getattr(ax, attr))
            def call(*args, **kwargs):
                for ax in self.stack.flat:
                    yield getattr(ax.axes, attr)(*args, **kwargs)
            return call
        def get_bboxes(self):
            return [
                    a.get_bbox() for a in self.stack.flat
                ]
        def get_bbox(self):
            bboxes = [
                a.get_bbox() for a in self.stack.flat
            ]

            bbox = [list(x) for x in bboxes[0]]
            for ((bl, br), (tl, tr)) in bboxes[1:]:
                bbox[0][0] = min(bbox[0][0], bl)
                bbox[0][1] = min(bbox[0][1], br)
                bbox[1][0] = max(bbox[1][0], tl)
                bbox[1][1] = max(bbox[1][1], tr)

            return bbox
        def get_padding(self):
            paddings = [
                [a.padding for a in ax]
                for ax in self.stack
            ]
            padding = [[0, 0], [0, 0]]
            for i,pad_list in enumerate(paddings):
                if i == 0:
                    padding[1][1] = max(p[1][1] for p in pad_list)
                elif i == len(paddings) - 1:
                    padding[1][0] = max(p[1][0] for p in pad_list)
                padding[0][0] = max([padding[0][0], pad_list[0][0][0]])
                padding[0][1] = max([padding[0][1], pad_list[-1][0][1]])

            return padding

        def set_facecolor(self, fg):
            pass

        def __iter__(self):
            return self.stack.flat
            # else:
            #     raise AttributeError("{} has no attribute {}".format(Graphics.__name__, attr))

    def initialize_figure_and_axes(self,
                      figure, axes, *,
                      nrows=None, ncols=None, graphics_class=None,
                      fig_kw=None, subplot_kw=None,
                      padding=None, spacings=None,
                      subimage_size=None, subimage_aspect_ratio=None,
                      **kw
                      ):
        """Initializes the subplots for the Graphics object

        :param figure:
        :type figure:
        :param axes:
        :type axes:
        :param args:
        :type args:
        :param kw:
        :type kw:
        :return: figure, axes
        :rtype: GraphicsBackend.Figure, GraphicsBackend.Figure.Axes
        """

        if 'image_size' not in kw:
            kw['image_size'] = subimage_size
        else:
            subimage_size = kw['image_size']
        if figure is None:
            padding = self._get_def_opt('padding', padding, {})
            # subimage_aspect_ratio = self._get_def_opt('aspect_ratio', subimage_aspect_ratio, {})
            spacings = self._get_def_opt('spacings', spacings, {})
            if subplot_kw is None:
                subplot_kw = {}
            if fig_kw is None:
                fig_kw = {}
            if 'figsize' not in fig_kw:
                try:
                    dw, dh = subimage_size
                except (ValueError, TypeError):
                    dw = dh = subimage_size
                w = ncols * dw
                h = nrows * dh
                if padding is not None:
                    pw, ph = padding
                    try:
                        pwx, pwy = pw
                    except (TypeError, ValueError):
                        pwx = pwy = pw
                    try:
                        phx, phy = ph
                    except (TypeError, ValueError):
                        phx = phy = ph

                    w += pwx + pwy
                    h += phx + phy
                if spacings is not None:
                    sw, sh = spacings
                    w += (ncols-1) * sw
                    h += (nrows-1) * sh
                fig_kw['figsize'] = (w/DPI_SCALING, h/DPI_SCALING)

            figure, axes = super().initialize_figure_and_axes(
                figure, axes, nrows=nrows, ncols=ncols, subplot_kw=subplot_kw,
                **fig_kw
            )


            if not isinstance(axes, (np.ndarray, list, tuple)):
                axes = [[axes]]
            elif not isinstance(axes[0], (np.ndarray, list, tuple)):
                if ncols == 1:
                    axes = [[ax] for ax in axes]
                else:
                    axes = [axes]

            axes = [[a for a in ax] for ax in axes]
            for i in range(nrows):
                for j in range(ncols):
                    axes[i][j] = graphics_class(figure=figure, axes=axes[i][j], managed=True, **kw)
            axes = self.GraphicsStack(self, axes)
        elif isinstance(figure, GraphicsGrid):
            axes = figure.axes
            figure = figure.figure

        if axes is None:
            axes = self.GraphicsStack(self, [
                graphics_class(
                    figure.create_axes(nrows, ncols, i),
                    managed=True,
                    **kw
                ) for i in range(nrows * ncols)
            ])

        return figure, axes

    def set_options(self,
                    padding=None,
                    spacings=None,
                    background=None,
                    colorbar=None,
                    figure_label=None,
                    **parent_opts
                    ):

        super().set_options(**parent_opts)

        self.image_size = None
        opts = (
            ('figure_label', figure_label),
            ('padding', padding),
            ('spacings', spacings),
            ('background', background),
            ('colorbar', colorbar)
        )
        for oname, oval in opts:
            oval = self._get_def_opt(oname, oval, {})
            if oval is not None:
                setattr(self, oname, oval)

    def __iter__(self):
        return iter(self.axes)
    def __getitem__(self, item):
        try:
            i, j = item
        except ValueError:
            return self.axes[item]
        else:
            return self.axes[i][j]
    def __setitem__(self, item, val):
        try:
            i, j = item
        except ValueError:
            if isinstance(val, GraphicsBase):
                self.axes[item] = val.change_figure(self.axes[item], image_size=val.image_size)
            else:
                self.axes[item] = val
        else:
            if isinstance(val, GraphicsBase):
                self.axes[i][j] = val.change_figure(self.axes[i][j], image_size=val.image_size)
            else:
                self.axes[i][j] = val
    def set_image(self, pos, val, **opts):
        pos = tuple(pos) if not isinstance(pos, int) else pos
        self.axes[pos] = val.change_figure(self.axes[pos], image_size=val.image_size, **opts)
        return self.axes[pos]

    # set size
    def calc_image_size(self):
        w=0; h=0
        for l in self.axes.stack:
            mh = 0; mw = 0
            for f in l:
                wh = f.image_size
                if wh is not None:
                    mw += wh[0]
                    if wh[1] > mh:
                        mh = wh[1]
            h += mh
            if w < mw:
                w = mw

        spacings = self.spacings
        if spacings is not None:
            sw, sh = spacings
            w += (self.shape[1] - 1) * sw
            h += (self.shape[0] - 1) * sh

        pad = self.padding
        if pad is not None:
            pw, ph = pad
            w += pw[0] + pw[1]
            h += ph[0] + ph[1]
        return w, h

    @property
    def image_size(self):
        s = self.calc_image_size()
        self._prop_manager.image_size = s
        return s
    @image_size.setter
    def image_size(self, value):
        s = self.calc_image_size()
        self._prop_manager.image_size = s

    @property
    def figure_label(self):
        return self._prop_manager.figure_label
    @figure_label.setter
    def figure_label(self, value):
        self._prop_manager.figure_label = value
    @property
    def padding(self):
        return self._prop_manager.padding
    @padding.setter
    def padding(self, value):
        self._prop_manager.padding = value
    @property
    def padding_left(self):
        return self._prop_manager.padding_left
    @padding_left.setter
    def padding_left(self, value):
        self._prop_manager.padding_left = value
    @property
    def padding_right(self):
        return self._prop_manager.padding_right
    @padding_right.setter
    def padding_right(self, value):
        self._prop_manager.padding_right = value
    @property
    def padding_top(self):
        return self._prop_manager.padding_top
    @padding_top.setter
    def padding_top(self, value):
        self._prop_manager.padding_top = value
    @property
    def padding_bottom(self):
        return self._prop_manager.padding_bottom
    @padding_bottom.setter
    def padding_bottom(self, value):
        self._prop_manager.padding_bottom = value

    @property
    def spacings(self):
        return self._prop_manager.spacings
    @spacings.setter
    def spacings(self, value):
        self._prop_manager.spacings = value

    @property
    def background(self):
        return self._prop_manager.background
    @background.setter
    def background(self, value):
        self._prop_manager.background = value

    @property
    def colorbar(self):
        return self._prop_manager.colorbar
    @colorbar.setter
    def colorbar(self, value):
        self._prop_manager.colorbar = value

    # def _prep_show(self):
    #     self.image_size = None
    #     if self.spacings is not None:
    #         self.spacings = self.spacings
    #     if self.padding is not None:
    #         self.padding = self.padding
    #     if self.tighten:
    #         self.figure.tight_layout()
        # super().prep_show()

    def _prep_show(self, parent=False):
        super()._prep_show(parent=parent)
        self.image_size = None
        if self.spacings is not None:
            self.spacings = self.spacings
        if self.padding is not None:
            self.padding = self.padding
        if self.tighten:
            self.figure.tight_layout()
    # def prep_show(self):
    #     if self.figure_parent is self:
    #         self._prep_show(parent=True)
    #         for c in self.children:
    #             if hasattr(c, '_prep_show'):
    #                 c._prep_show(parent=False)
    #             else:
    #                 c.prep_show()
    #     else:
    #         self.figure_parent.prep_show()
    #     return self

    # def show(self, **kwargs):
    #     super().show(**kwargs)




"""
Defines a helper class Styled to make it easier to style plots and stuff and a ThemeManager to handle all that shiz
"""
import contextlib
from collections import deque
from .Backends import GraphicsBackend

__all__ = [
    "Styled",
    "ThemeManager",
    "PlotLegend"
]


class Styled:
    """
    Simple styling class
    """
    def __init__(self, *str, **opts):
        if len(str) == 2 and isinstance(str[1], dict):
            opts = dict(str[1], **opts)
            str = [str[0]]
        self.val = str
        self.opts = opts
    @classmethod
    def could_be(cls, data):
        return isinstance(data, tuple) and len(data) == 2 and isinstance(data[1], dict)
    @classmethod
    def construct(cls, data):
        return cls(data[0], **data[1])

class PlotLegend(list):
    known_styles = {"handles", "labels", "loc", "bbox_to_anchor", "ncol", "prop", "fontsize",
                    "labelcolor", "numpoints", "scatterpoints", "scatteryoffsets", "markerscale",
                    "markerfirst", "frameon", "fancybox", "shadow", "framealpha", "facecolor",
                    "edgecolor", "mode", "bbox_transform", "title", "title_fontproperties",
                    "title_fontsize", "borderpad", "labelspacing", "handlelength", "handleheight",
                    "handletextpad", "borderaxespad", "columnspacing", "handler_map"}
    default_styles={'frameon':False}
    def __init__(self, components, **styles):
        if isinstance(components, type(self)):
            self.__init__(list(self), **self.opts)
        else:
            super().__init__(components)
            self.check_styles(styles)
            for d in self.default_styles - styles.keys():
                styles[d] = self.default_styles[d]
            self.opts = styles
    @classmethod
    def check_styles(cls, styles):
        unkown = styles.keys() - cls.known_styles
        if len(unkown) > 0:
            raise ValueError("styles {} not known for {}".format(unkown, cls.__name__))
    @classmethod
    def could_be_legend(cls, bits):
        if isinstance(bits, (str, int, float)):
            return False
        try:
            iter(bits)
        except TypeError:
            return False
        return True
    @classmethod
    def construct(cls, bits):
        if isinstance(bits, cls):
            return bits
        elif len(bits) == 2 and hasattr(bits[1], 'items') and not hasattr(bits[0], 'items'):
            bits, opts = bits
        else:
            opts = {}
        bits = [b if not hasattr(b, 'items') else cls.canonicalize_bit(**b) for b in bits]
        return cls(bits, **opts)
    @classmethod
    def construct_line_marker(cls, lw=4, **opts):
        from matplotlib.lines import Line2D
        return Line2D([0], [0], lw=lw, **opts)
    @classmethod
    def construct_dot_marker(cls, **opts):
        from matplotlib.patches import Patch
        return Patch(**opts)
    @classmethod
    def load_constructors(cls):
        return {
        'line':cls.construct_line_marker,
        'dot':cls.construct_dot_marker
    }
    marker_synonyms={'-':'line', '.':'dot'}
    @classmethod
    def canonicalize_bit(cls, marker='-', **opts):
        if isinstance(marker, str) and marker in cls.marker_synonyms:
            marker = cls.marker_synonyms[marker]
        if isinstance(marker, str):
            return cls.load_constructors()[marker](**opts)
        else:
            return marker(**opts)
    def __repr__(self):
        return "{}({}, {})".format(
            type(self).__name__,
            super().__repr__(),
            self.opts
        )

class cycler:
    # transparent mimic of mpl cycler
    def __init__(self, **style_cycles):
        self.opts = style_cycles
        self.keys = list(style_cycles.keys())
        self.lens = [len(style_cycles[k]) for k in self.keys]
        self.inds = [0]*len(style_cycles)
    def __next__(self):
        vals = {
            k:self.opts[k][i]
            for k,i in zip(self.keys, self.inds)
        }
        self.inds = [(i+1)%l for i,l in zip(self.inds, self.lens)]
        return vals

class ThemeManager:
    """
    Simple manager class for plugging into themes in a semi-background agnostic way
    """
    extra_themes = {
        'mccoy': (
            ([],),
            {
                'axes': {
                    'prop_cycle':{'color': ['001C7F', '017517', '8C0900', '7600A1', 'B8860B', '006374']},
                    'labelsize':13
                },
                'patch': {'facecolor': '001C7F'},
                'xtick': {'labelsize': 13},
                'ytick': {'labelsize': 13},
                'padding': 50,
                'aspect_ratio': 'auto'
            }
        )

    }
    _resolved_theme_cache = {

    }
    def __init__(self, *theme_names, backend=None, graphics_styles=None, **extra_styles):
        self.main_theme_names = theme_names
        self.extra_styles = extra_styles
        self.graphics_styles = graphics_styles
        if backend is None: backend = GraphicsBackend.lookup('matplotlib')
        self.backend = backend
        self.context_manager = None
    @classmethod
    def from_spec(cls, theme, backend=None):
        if theme is None:
            return NoThemeManager()
        if isinstance(theme, str):
            theme = [theme]
        elif isinstance(theme, dict):
            theme = [(), theme]
        if len(theme) > 0:
            try:
                theme_names, theme_properties = theme
            except ValueError:
                theme_names = theme[0]
                theme_properties = {}
            if isinstance(theme_names, dict):
                theme_properties = theme_names
                theme_names = []
            elif isinstance(theme_names, str):
                theme_names = [theme_names]
        else:
            theme_names = []
            theme_properties = {}
        return cls(*theme_names, backend=backend, **theme_properties)
    def _test_rcparam(self, k):
        return '.' in k
    def __enter__(self):
        theme = self.resolve_theme(None, *self.main_theme_names, **self.extra_styles)
        name_list = self.validate_theme(*theme)
        # name_list = list(theme[0])
        # opts = {k:v for k,v in theme[1].items() if self._test_rcparam(k)}

        self.context_manager = self.backend.theme_context(name_list, theme[1])
        return self.context_manager.__enter__()
        # don't currently support any other backends...
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.context_manager is not None:
            self.context_manager.__exit__(exc_type, exc_val, exc_tb)
            self.context_manager = None
    @property
    def theme(self):
        return self.resolve_theme(None, *self.main_theme_names, **self.extra_styles)
    @classmethod
    def add_theme(self, theme_name, *base_theme, **extra_styles):
        """
        Adds a theme to the extra themes dict. At some future date we'll
        want to make it so that this does a level of validation, too.
        :param theme_name:
        :type theme_name:
        :param base_theme:
        :type base_theme:
        :param extra_styles:
        :type extra_styles:
        :return:
        :rtype:
        """
        self.extra_themes[theme_name] = (base_theme, extra_styles)
    @classmethod
    def resolve_theme(self, theme_name, *base_themes, **extra_styles):
        """
        Resolves a theme so that it only uses strings for built-in styles
        :return:
        :rtype:
        """
        if theme_name is not None:
            if theme_name in self._resolved_theme_cache:
                themes, styles = self._resolved_theme_cache[theme_name]
            elif theme_name in self.extra_themes:
                # recursively resolve the theme
                bases, extras = self.extra_themes[theme_name]
                if isinstance(bases, str):
                    bases = [bases]
                theme_stack = deque()
                style_stack = deque()
                for theme_list in bases:
                    if isinstance(theme_list, str):
                        theme_list = [theme_list]
                    remainder_themes = []
                    for theme in theme_list:
                        if theme in self.extra_themes:
                            t, s = self.resolve_theme(theme_name)
                            theme_stack.appendleft(t)
                            style_stack.append(s)
                        else:
                            remainder_themes.append(theme)
                    if len(remainder_themes) > 0:
                        theme_stack.appendleft([remainder_themes])
                themes = tuple(x for y in theme_stack for x in y)
                styles = {}
                for s in style_stack:
                    styles.update(s)
                styles.update(extras)
                self._resolved_theme_cache[theme_name] = [themes, styles]
            else:
                themes = (theme_name,)
                styles = {}
        else:
            themes = ()
            styles = {}

        for name_list in base_themes:
            if isinstance(name_list, str):
                name_list = [name_list]
            remainder_themes = []
            for b in name_list:
                if b in self.extra_themes:
                    t, s = self.resolve_theme(b)
                    themes = tuple(t) + themes
                    styles.update(s)
                else:
                    remainder_themes.append(b)
            if len(remainder_themes) > 0:
                themes = (remainder_themes,) + themes

        styles.update(extra_styles)

        return [themes, styles]
    def validate_theme(self, theme_names, theme_styless):
        valid_names = set(self.backend_themes)
        resolved_names = []
        for k in theme_names:
            if isinstance(k, str):
                if k not in valid_names:
                    raise ValueError("{}.{}: theme '{}' isn't in supported set ({})".format(
                        type(self).__name__,
                        'validate_theme',
                        k,
                        valid_names
                    ))
                else:
                    resolved_names.append(k)
            else:
                for altname in k:
                    if altname in valid_names:
                        resolved_names.append(altname)
                        break
                else:
                    raise ValueError("{}.{}: no theme in '{}' isn't in supported set ({})".format(
                        type(self).__name__,
                        'validate_theme',
                        k,
                        valid_names
                    ))
        return resolved_names

    @property
    def backend_themes(self):
        theme_names = self.backend.get_available_themes()
        return tuple(theme_names)
    @property
    def theme_names(self):
        return self.backend_themes + tuple(self.extra_themes.keys())

class NoThemeManager:
    """
    Does nothing but makes code consistent
    """
    def __enter__(self):
        pass
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import scipy.special as funcs, numpy as np

data = {}

def harmonic(r, *, re=1, freq=1, mass=1, deriv_order=None):
    r = np.asanyarray(r)
    c = r - re
    k = (mass*(freq**2))
    vals = k/2 * c**2
    if deriv_order is not None:
        vals = [vals]
        for n in range(1, deriv_order+1):
            if n == 1:
                vals.append(k*c)
            elif n == 2:
                if not isinstance(k, np.ndarray):
                    k = np.full(r.shape, k)
                vals.append(k)
            else:
                vals.append(np.zeros(r.shape))
    return vals
data['HarmonicPotential'] = {
        "Name":"HarmonicPotential",
        "Parameters":{
            "re":"Equilibrium bond length",
            "freq":"Frequency of the oscillator",
            "mass":"Mass of the oscillator"
        },
        "Defaults":{
            "re":1,
            "freq":1,
            "mass":1
        },
        "Function": harmonic
    }

def morse(r, *, re=1, De=1, alpha=1, deriv_order=None):
    r = np.asanyarray(r)
    c = alpha * (r - re)
    vals = De * (1 - np.exp(-c)) ** 2
    if deriv_order is not None:
        vals = [vals]
        for n in range(1, deriv_order+1):
            d = (-1) **(n + 1) *( 2 * De * alpha**n) * np.exp(-2 * c) * (np.exp(c) - (2**(n - 1)))
            vals.append(d)
    return vals
data["MorsePotential"] = {
        "Name":"MorsePotential",
        "Parameters":{
            "re":"Equilibrium bond length",
            "De":"Dissociation energy",
            "alpha":"Width of the well"
        },
        "Defaults":{
            "re":1,
            "De":1,
            "alpha":1
        },
        "Function": morse
    }